# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2017-2019 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "06/26/2019"


import unittest
import tempfile
import shutil
from est.core.utils import DownloadDataset
from ..process import exec_
from pypushflow.representation.scheme.ows_parser import OwsParser
import os
from est.io.utils.information import InputInformation
from silx.io.url import DataUrl

try:
    import PyMca5
except ImportError:
    has_pymca = False
else:
    has_pymca = True
    from PyMca5.PyMcaDataDir import PYMCA_DATA_DIR


@unittest.skipIf(has_pymca is False, "PyMca5 is not installed")
class TestWorkflowFromOwsFile(unittest.TestCase):
    """test construction of XAS object"""

    def setUp(self):
        self.outputdir = tempfile.mkdtemp()
        file_ = "pymca_workflow_2.ows"
        DownloadDataset(dataset=file_, output_folder=self.outputdir, timeout=10.0)
        self.orange_file = os.path.join(self.outputdir, file_)
        self.input_file1 = os.path.join(PYMCA_DATA_DIR, "EXAFS_Cu.dat")
        self.output_file = os.path.join(self.outputdir, "output.h5")

    def tearDown(self):
        shutil.rmtree(self.outputdir)

    def testPyMcaWorkflow(self):
        """Test regarding the instantiation of the pymcaXAS"""
        exec_(
            scheme=OwsParser.scheme_load(self.orange_file, load_handlers=True),
            input_information=InputInformation(
                spectra_url=DataUrl(
                    file_path=self.input_file1, scheme="PyMca", data_path="Column 1"
                ),
                channel_url=DataUrl(
                    file_path=self.input_file1, scheme="PyMca", data_path="Column 2"
                ),
            ),
            output_=self.output_file,
        )
        self.assertTrue(os.path.exists(self.output_file))
        # TODO: check the values in the output file


def suite():
    test_suite = unittest.TestSuite()
    for ui in (TestWorkflowFromOwsFile,):
        test_suite.addTest(unittest.defaultTestLoader.loadTestsFromTestCase(ui))
    return test_suite


if __name__ == "__main__":
    unittest.main(defaultTest="suite")



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "06/11/2019"

from est.core.types import XASObject, Spectrum
from .process import Process
import logging

_logger = logging.getLogger(__name__)


def _xas_set_config(xas_obj):
    """
    Copy process configuration to object

    :param xas_obj: object containing the configuration and spectra to process
    :type: Union[:class:`.XASObject`, dict]
    :return: spectra dict
    :rtype: :class:`.XASObject`
    """
    xas_set_config = _SetConfigProcess()
    return xas_set_config.process(xas_obj=xas_obj)


class _SetConfigProcess(Process):
    def __init__(self):
        Process.__init__(self, name="set_config")

    def process(self, xas_obj):
        """

        :param xas_obj: object containing the configuration and spectra to process
        :type: Union[:class:`.XASObject`, dict]
        :return: spectra dict
        :rtype: :class:`.XASObject`
        """
        assert xas_obj is not None
        _xas_obj = self.getXasObject(xas_obj=xas_obj)
        conf_obj = _xas_obj.configuration
        conf_process = self.getConfiguration()
        for key, value in conf_process.items():
            conf_obj[key] = value
        _xas_obj.configuration = conf_obj
        return _xas_obj



import argparse
import sys
from pypushflow.Workflow import ProcessableWorkflow
from pypushflow.representation.scheme.ows_parser import OwsParser
import logging
import signal
from pypushflow.representation.scheme.scheme import Scheme
from silx.io.url import DataUrl
from est.units import ur
from est.io.utils.information import InputInformation
from est.io.utils.information import SpecInputInformation
from .utils import get_unit
from .utils import get_url
from .utils import convert_spectra_dims
from .utils import get_xas_obj

try:
    import h5py

    has_h5py = True
except:
    has_h5py = False
from typing import Union

logging.basicConfig(level=logging.DEBUG)
_logger = logging.getLogger(__name__)


def _insert_input_in_scheme(scheme, input_information):
    """update 'starting' node properties to include the provided input"""
    # monkey patch the input file for start nodes if an input is given
    for node in scheme.nodes:
        if node.properties and "_spec_file_setting" in node.properties:
            if input_information.is_spec_input():
                node.properties["_spec_file_setting"] = input_information.spec_info
            if input_information.spectra_url is not None:
                node.properties[
                    "_spectra_url_setting"
                ] = input_information.spectra_url.path()
            if input_information.dimensions is not None:
                node.properties["_dimensions_setting"] = input_information.dimensions
            if input_information.channel_url is not None:
                node.properties[
                    "_energy_url_setting"
                ] = input_information.channel_url.path()
            if input_information.configuration_url is not None:
                node.properties[
                    "_configuration_url_setting"
                ] = input_information.configuration_url.path()


def _insert_output_in_scheme(scheme, output_):
    """update 'starting' node properties to include the provided input"""
    found_output = False
    # monkey patch the input file for start nodes if an input is given
    for node in scheme.nodes:
        if node.properties and "_output_file_setting" in node.properties:
            node.properties["_output_file_setting"] = output_
            found_output = True
    if not found_output:
        _logger.warning(
            "No node for processing output found. output "
            "information provided will be ignored"
        )


def exec_(
    scheme: Scheme,
    input_information: InputInformation,
    output_: Union[str, None, dict] = None,
    timeout: Union[int, None] = None,
):

    if not input_information.is_valid():
        raise ValueError("You cannot provide an input file and input urls")

    _insert_input_in_scheme(input_information=input_information, scheme=scheme)
    if output_ is not None:
        _insert_output_in_scheme(scheme=scheme, output_=output_)

    workflow = ProcessableWorkflow(scheme=scheme)

    # add SIGINT capture
    def signal_handler(sig, frame):
        _logger.warning("stop workflow execution on user request")
        workflow._end_actor.join(0)
        sys.exit(0)

    signal.signal(signal.SIGINT, signal_handler)

    xas_obj = get_xas_obj(input_information=input_information)

    workflow._start_actor.trigger(("data", xas_obj.to_dict()))
    workflow._end_actor.join(timeout)
    res = workflow._end_actor.out_data
    title = scheme.title or "unknow"
    _logger.info(
        "workflow '{}' completed with {}".format(title, str(input_information))
    )
    return res


def main(argv):
    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument(
        "workflow_file",
        help="Workflow file providing the workflow description (.ows, .xml)",
    )
    # single file input option
    parser.add_argument(
        "-i",
        "--input",
        dest="input_",
        default=None,
        help="Input of the workflow. Should be a path to a file",
    )
    # input url option
    parser.add_argument(
        "--input-spectra",
        "--spectra",
        dest="input_spectra",
        default=None,
        help="Input spectra url",
    )
    parser.add_argument(
        "--input-spectra-dims",
        "--spectra-dims",
        dest="input_spectra_dims",
        default=None,
        help="Input spectra dimension. Should be a tuple of three values: "
        "(X,Y,channel). If None will take the default dimension "
        "according to the input type.",
    )
    parser.add_argument(
        "--input-energy",
        "--input-channel",
        "--channel",
        dest="input_channel",
        default=None,
        help="Input channel url (usually energy)",
    )
    parser.add_argument(
        "--input-configuration",
        "--configuration",
        dest="input_configuration",
        default=None,
        help="Input configuration url",
    )
    parser.add_argument(
        "--input-energy-unit",
        "--energy-unit",
        dest="input_energy_unit",
        default="eV",
        help="energy unit",
    )
    parser.add_argument(
        "--input-dimensions",
        "--dimensions",
        dest="input_dimensions",
        default="None",
        help="dimension of the input as (Z,Y,X) for example."
        "If None will take default unit according to the input type",
    )
    # I0, I1, I2 & mu_ref
    parser.add_argument(
        "--input-I0",
        "--I0",
        dest="input_I0",
        default="None",
        help="url to I0",
    )
    parser.add_argument(
        "--input-I1",
        "--I1",
        dest="input_I1",
        default="None",
        help="url to I1",
    )
    parser.add_argument(
        "--input-I2",
        "--I2",
        dest="input_I2",
        default="None",
        help="url to I2",
    )
    parser.add_argument(
        "--input-mu-ref",
        "--mu-ref",
        dest="input_mu_ref",
        default="None",
        help="url to mu_ref",
    )
    # spec file specific inputs
    parser.add_argument(
        "--input-energy-col-name",
        "--energy-col-name",
        dest="input_energy_col_name",
        default=None,
        help="Provide name of the energy column for spec file",
    )
    parser.add_argument(
        "--input-abs-col-name",
        "--abs-col-name",
        dest="input_abs_col_name",
        default=None,
        help="Provide name of the absorption column for spec file",
    )
    parser.add_argument(
        "--input-monitor-col-name",
        "--monitor-col-name",
        dest="input_monitor_col_name",
        default=None,
        help="Provide name of the monitor column for spec file",
    )
    parser.add_argument(
        "--input-scan-title",
        "--scan-title",
        dest="input_scan_title_name",
        default=None,
        help="Provide scan title name to consider",
    )
    # output option
    parser.add_argument(
        "-o",
        "--output",
        dest="output_",
        default=None,
        help="Output file of the workflow. Require at most one "
        "instance of XASOutputOW",
    )
    options = parser.parse_args(argv[1:])

    scheme = OwsParser.scheme_load(options.workflow_file, load_handlers=True)
    input_information = InputInformation(
        spectra_url=get_url(options.input_spectra),
        channel_url=get_url(options.input_channel),
        dimensions=convert_spectra_dims(options.input_spectra_dims),
        config_url=get_url(options.input_configuration),
        energy_unit=get_unit(options.input_energy_unit),
        spec_input=SpecInputInformation(
            options.input_,
            options.input_energy_col_name,
            options.input_abs_col_name,
            options.input_monitor_col_name,
            options.input_scan_title_name,
        ),
    )
    input_information.I0 = get_url(options.input_I0)
    input_information.I1 = get_url(options.input_I1)
    input_information.I2 = get_url(options.input_I2)
    input_information.mu_ref = get_url(options.input_mu_ref)
    exec_(scheme=scheme, input_information=input_information)


if __name__ == "__main__":
    main(sys.argv)



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "18/01/2021"

from est.core.types import XASObject
from est.core.types import Spectrum
from est.core.process.process import _input_desc
from est.core.process.process import _output_desc
from .process import Process
from .process import _NexusSpectrumDef
from .process import _NexusDatasetDef
import scipy.signal
import logging
import numpy
import pkg_resources
import multiprocessing
import functools
from typing import Union

_logger = logging.getLogger(__name__)


_USE_MULTIPROCESSING_POOL = False
# note: we cannot use multiprocessing pool with pypushflow for now.


def process_noise_savgol(
    spectrum,
    configuration,
    overwrite=True,
    callbacks=None,
    output=None,
    output_dict=None,
):
    """

    :param spectrum: spectrum to process
    :type: :class:`.Spectrum`
    :param configuration: configuration of the pymca normalization
    :type: dict
    :param overwrite: False if we want to return a new Spectrum instance
    :type: bool
    :param callbacks: callback to execute.
    :param output: list to store the result, needed for pool processing
    :type: multiprocessing.manager.list
    :param output_dict: key is input spectrum, value is index in the output
                        list.
    :type: dict
    :return: processed spectrum
    :rtype: tuple (configuration, spectrum)
    """
    _logger.debug(
        "start noise with Savitsky-Golay on spectrum (%s, %s)"
        % (spectrum.x, spectrum.y)
    )
    if "noise" in configuration:
        configuration = configuration["noise"]
    if not "window_size" in configuration:
        raise ValueError("`window_size` should be specify. Missing in configuration")
    else:
        window_size = configuration["window_size"]
    if not "polynomial_order" in configuration:
        raise ValueError(
            "`polynomial_order` should be specify. Missing in configuration"
        )
    else:
        polynomial_order = configuration["polynomial_order"]
    if not "e_min" in configuration:
        e_min = None
    else:
        e_min = configuration["e_min"]
    if not "e_max" in configuration:
        e_max = None
    else:
        e_max = configuration["e_max"]

    if e_min in (None, -1):
        e_min = spectrum.energy.min() - 1
    if e_max in (None, -1):
        e_max = spectrum.energy.max() + 1

    mask = (spectrum.energy > e_min) & (spectrum.energy < (e_max))
    noise_savgol_energy = spectrum.energy[mask]
    raw_mu = spectrum.mu[mask]
    smooth_spectrum = scipy.signal.savgol_filter(raw_mu, window_size, polynomial_order)
    noise = numpy.absolute(raw_mu - smooth_spectrum)
    spectrum.noise_savgol = noise
    if hasattr(spectrum, "edge_step"):
        spectrum.norm_noise_savgol = numpy.mean(noise) / spectrum.edge_step
    else:
        spectrum.norm_noise_savgol = None
        _logger.warning(
            "Unable to compute Normalized edge."
            "Normalization (pre-edge) should be run first"
        )
    spectrum.noise_savgol_energy = noise_savgol_energy

    if callbacks:
        for callback in callbacks:
            callback()

    if overwrite:
        spectrum_ = spectrum
    else:
        spectrum_ = Spectrum()
        spectrum_.update(spectrum)

    if output is not None:
        assert output_dict is not None
        output[output_dict[spectrum]] = spectrum_
    return configuration, spectrum_


def xas_noise(xas_obj):
    """
    apply roi on the XASObject.spectra

    :param xas_obj: object containing the configuration and spectra to process
    :type: Union[:class:`.XASObject`, dict]
    :return: spectra dict
    :rtype: :class:`.XASObject`
    """
    xas_noise_process = NoiseProcess()
    return xas_noise_process.process(xas_obj=xas_obj)


class NoiseProcess(Process):

    inputs = [
        _input_desc(name="xas_obj", type=XASObject, handler="process", doc=""),
    ]

    outputs = [
        _output_desc(name="xas_obj", type=XASObject, doc=""),
    ]

    def __init__(self):
        Process.__init__(self, name="noise")
        self._window_size = 5
        self._polynomial_order = 2

    def set_properties(self, properties: dict):
        if "noise" in properties:
            properties = properties["noise"]
        self._settings = properties
        if "window_size" in properties:
            self._window_size = properties["window_size"]
        if "polynomial_order" in properties:
            self._polynomial_order = properties["polynomial_order"]

    def process(self, xas_obj: Union[dict, XASObject]):
        """

        :param xas_obj: object containing the configuration and spectra to process
        :type: Union[:class:`.XASObject`, dict]
        :return: spectra dict
        :rtype: :class:`.XASObject`
        """
        assert xas_obj is not None
        _xas_obj = self.getXasObject(xas_obj=xas_obj)

        if self._settings:
            _xas_obj.configuration["noise"] = self._settings

        self._advancement.reset(max_=_xas_obj.n_spectrum)
        self._advancement.startProcess()
        self._pool_process(xas_obj=_xas_obj)
        self._advancement.endProcess()
        self.register_process(
            _xas_obj,
            data_keys=(
                _NexusDatasetDef("norm_noise_savgol"),
                _NexusDatasetDef("noise_savgol", units="raw data noise"),
                _NexusDatasetDef("noise_savgol_energy", units="eV"),
                _NexusDatasetDef("edge_step"),
            ),
            plots=(
                _NexusSpectrumDef(
                    signal="noise_savgol",
                    axes=("noise_savgol_energy",),
                    auxiliary_signals=None,
                    silx_style={"signal_scale_type": "log"},
                    title="noise",
                ),
            ),
        )
        return _xas_obj

    def _pool_process(self, xas_obj):
        assert isinstance(xas_obj, XASObject)
        if not _USE_MULTIPROCESSING_POOL:
            for spectrum in xas_obj.spectra:
                process_noise_savgol(
                    spectrum=spectrum,
                    configuration=xas_obj.configuration,
                    callbacks=self.callbacks,
                    overwrite=True,
                )
        else:
            from multiprocessing import Manager

            manager = Manager()
            output_dict = {}
            res_list = manager.list()
            for i_spect, spect in enumerate(xas_obj.spectra):
                res_list.append(None)
                output_dict[spect] = i_spect

            with multiprocessing.Pool(1) as p:
                partial_ = functools.partial(
                    process_noise_savgol,
                    configuration=xas_obj.configuration,
                    callbacks=self.callbacks,
                    overwrite=False,
                    output=res_list,
                    output_dict=output_dict,
                )
                p.map(partial_, xas_obj.spectra)

            # then update local spectrum
            for spectrum, res in zip(xas_obj.spectra, res_list):
                spectrum.update(res)

    def definition(self):
        return "noise using Savitsky-Golay algorithm"

    def program_version(self):
        return pkg_resources.get_distribution("est").version

    def program_name(self):
        return "noise_savgol"

    __call__ = process



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "06/07/2019"


import functools
import logging
from Orange.widgets import gui
from Orange.widgets.settings import Setting
from Orange.widgets.widget import OWWidget
from Orange.widgets.widget import Input, Output
import Orange.data
from silx.gui import qt
from silx.gui.plot import LegendSelector
import est.core.process.larch.pre_edge
from orangecontrib.est.process import _ProcessForOrangeMixIn
from orangecontrib.est.process import ProcessRunnable
from est.core.types import XASObject
from est.gui.XasObjectViewer import XasObjectViewer, ViewType
from est.gui.XasObjectViewer import (
    _plot_raw,
    _plot_edge,
    _plot_norm,
    _plot_norm_area,
    _plot_post_edge,
    _plot_pre_edge,
)
from est.gui.larch.pre_edge import _MPreEdgeParameters
from orangecontrib.est.progress import QProgress
from orangecontrib.est.utils import Converter
from orangecontrib.est.widgets.container import _ParameterWindowContainer

_logger = logging.getLogger(__file__)

_USE_THREAD = False


class _PreEdgeWindow(qt.QMainWindow):
    def __init__(self, parent=None):
        qt.QMainWindow.__init__(self, parent)

        # xas object viewer
        mapKeys = ["mu", "norm", "norm_area"]
        self.xasObjViewer = XasObjectViewer(mapKeys=mapKeys)
        self.xasObjViewer._spectrumViews[0]._plotWidget.getXAxis().setLabel(
            "Energy (eV)"
        )
        self.setCentralWidget(self.xasObjViewer)
        self._parametersWindow = _ParameterWindowContainer(
            parent=self, parametersWindow=_MPreEdgeParameters
        )
        dockWidget = qt.QDockWidget(parent=self)

        # parameters window
        dockWidget.setWidget(self._parametersWindow)
        self.addDockWidget(qt.Qt.RightDockWidgetArea, dockWidget)
        dockWidget.setAllowedAreas(qt.Qt.RightDockWidgetArea | qt.Qt.LeftDockWidgetArea)
        dockWidget.setFeatures(qt.QDockWidget.NoDockWidgetFeatures)

        # legend selector
        self.legendDockWidget = LegendSelector.LegendsDockWidget(
            parent=self, plot=self.xasObjViewer._spectrumViews[0]._plotWidget
        )
        self.legendDockWidget.setAllowedAreas(
            qt.Qt.RightDockWidgetArea | qt.Qt.LeftDockWidgetArea
        )
        self.legendDockWidget.setFeatures(qt.QDockWidget.NoDockWidgetFeatures)
        self.addDockWidget(qt.Qt.RightDockWidgetArea, self.legendDockWidget)

        # volume key selection
        self.addDockWidget(
            qt.Qt.RightDockWidgetArea, self.xasObjViewer._mapView.keySelectionDocker
        )

        # plot settings
        for ope in (
            _plot_raw,
            _plot_norm,
            _plot_norm_area,
            _plot_pre_edge,
            _plot_post_edge,
            _plot_edge,
        ):
            self.xasObjViewer._spectrumViews[0].addCurveOperation(ope)

        self.setWindowFlags(qt.Qt.Widget)

        # connect signal / slot
        self.xasObjViewer.viewTypeChanged.connect(self._updateLegendView)

        # set up
        self._updateLegendView()

    def _updateLegendView(self):
        index, viewType = self.xasObjViewer.getViewType()
        self.legendDockWidget.setVisible(viewType is ViewType.spectrum)
        self.xasObjViewer._mapView.keySelectionDocker.setVisible(
            viewType is ViewType.map
        )

    def getNCurves(self):
        return len(self.xasObjViewer._spectrumViews[0]._plotWidget.getAllCurves())


class PreEdgeOW(_ProcessForOrangeMixIn, OWWidget):
    """
    Widget used for signal extraction
    """

    name = "pre edge"
    id = "orange.widgets.est.larch.pre_edge.PreEdgeOW"
    description = """
    pre edge subtraction, normalization for XAFS
    This performs a number of steps:
       1. determine E0 (if not supplied) from max of deriv(mu)
       2. fit a line of polynomial to the region below the edge
       3. fit a polynomial to the region above the edge
       4. extrapolate the two curves to E0 to determine the edge jump
       5. estimate area from emin_area to norm2, to get norm_area
    """
    icon = "icons/pre_edge.png"
    priority = 2
    category = "esrfWidgets"
    keywords = ["spectroscopy", "preedge", "pre", "edge", "normalization"]

    want_main_area = True
    resizing_enabled = True
    allows_cycle = False

    process_function = est.core.process.larch.pre_edge.Larch_pre_edge

    _larchSettings = Setting(dict())
    """Store the configuration of the larch configuration"""

    class Inputs:
        xas_obj = Input("xas_obj", XASObject, default=True)
        # simple compatibility for some Orange widget and especialy the
        # 'spectroscopy add-on'
        data_table = Input("Data", Orange.data.Table)

    class Outputs:
        res_xas_obj = Output("xas_obj", XASObject)
        # by default we want to avoid sending 'Orange.data.Table' to avoid
        # loosing the XASObject flow process and results.

    def __init__(self):
        super().__init__()
        self._latest_xas_obj = None
        self._window = _PreEdgeWindow(parent=self)
        layout = gui.vBox(self.mainArea, "pre edge").layout()
        layout.addWidget(self._window)
        self._window.xasObjViewer.setWindowTitle("spectra")

        # manage settings
        if self._larchSettings != dict():
            self._window._parametersWindow.setParameters(self._larchSettings)

        # connect signals / slots
        self._window._parametersWindow.sigChanged.connect(self._updateProcess)

        # required to display advancement
        if _USE_THREAD is False:
            self._advancement = QProgress("pre_edge")
            self._advancement.sigProgress.connect(self._setProgressValue)

        # set up (insure settings will be store
        self._update_settings()

    def _updateProcess(self):
        self._update_settings()
        if self._latest_xas_obj:
            self.process(self._latest_xas_obj)

    def _update_settings(self):
        self._larchSettings = self._window._parametersWindow.getParameters()

    @Inputs.data_table
    def processFrmDataTable(self, data_table):
        if data_table is None:
            return
        self.process(Converter.toXASObject(data_table=data_table))

    @Inputs.xas_obj
    def process(self, xas_obj):
        if xas_obj is None:
            return

        if not self._canProcess():
            _logger.warning(
                "There is some processing on going already, will"
                "not process the new dataset"
            )

        self._latest_xas_obj = xas_obj

        # e0 can be defined from the flow previously
        if "e0" in xas_obj.configuration:
            conf = {"e0": xas_obj.configuration["e0"]}
            self._window._parametersWindow.setParameters(conf)
        self._startProcess()

        # setup the normalization process
        if _USE_THREAD is True:
            # larch is using sqlite3 for database. access to the database
            # require a 'connection' and 'cursor'. theoretically we should be
            # able to access it from a single thread using something like:
            #
            # import sqlite3
            # from larch.xray import xraydb
            # import os
            #
            # dbname = 'xraydata.db'
            # parent, _ = os.path.split(xraydb.__file__)
            # file_ = os.path.join(parent, dbname)
            # if not os.path.exists(file_):
            #     raise IOError('%s does not exists' % file_)
            # self.connection = sqlite3.connect(file_)
            # cursor = self.connection.cursor()
            # self._processingThread = PreEdgeThread(advancement=self._advancement,
            #                                        args=(self.connection, cursor))
            # working until at one point (in time) the database is closed by larch
            # or there is some concurrency failling ?
            # so fo rnow we are making the processing in the main thread that
            # larch is happy about and splitting it in a way to avoid gui
            # freeze
            process_obj = QLarch_pre_edge()
            process_obj._advancement.sigProgress.connect(self._setProgressValue)
            process_obj.set_properties(
                {"_larchSettings": self._window._parametersWindow.getParameters()}
            )
            # update the processing thread
            thread = self.getProcessingThread()
            thread.init(process_obj=process_obj, xas_obj=self._latest_xas_obj)
            self._callback_finish = functools.partial(
                self._endProcess, self._latest_xas_obj
            )
            thread.finished.connect(self._callback_finish)
            # start processing
            thread.start(priority=qt.QThread.LowPriority)
        else:
            # manage advancement
            self._advancement.setAdvancement(0)
            self._advancement.setMaxSpectrum(self._latest_xas_obj.n_spectrum)
            process_obj = est.core.process.larch.pre_edge.Larch_pre_edge()
            process_obj.advancement = self._advancement
            process_obj.set_properties(
                {"_larchSettings": self._window._parametersWindow.getParameters()}
            )
            # hack (reason explained in the thread part)
            # to avoid gui freeze, processevents after each spectrum process
            process_obj.addCallback(qt.QApplication.instance().processEvents)
            process_obj.addCallback(self._advancement.increaseAdvancement)
            process_obj.process(self._latest_xas_obj)
            self._callback_finish = None
            self._endProcess(self._latest_xas_obj)


class QLarch_pre_edge(est.core.process.larch.pre_edge.Larch_pre_edge):
    """
    Normalization able to give advancement using qt.Signal and QThreadPool
    """

    def __init__(self):
        est.core.process.larch.pre_edge.Larch_pre_edge.__init__(self)
        self._advancement = QProgress("pre_edge")

    def _pool_process(self, xas_obj):
        self.pool = qt.QThreadPool()
        self.pool.setMaxThreadCount(5)
        for spectrum in xas_obj.spectra:
            runnable = ProcessRunnable(
                fct=est.core.process.larch.pre_edge.process_spectr_pre_edge,
                spectrum=spectrum,
                configuration=xas_obj.configuration,
                callback=self._advancement.increaseAdvancement,
            )
            self.pool.start(runnable)
        self.pool.waitForDone()



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2017-2019 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "06/26/2019"


import numpy
import os
import unittest
import tempfile
import h5py
import shutil
from est.core.types import Spectrum, XASObject, Dim
from est.core.utils import spectra as spectra_utils
from est.core.io import read as read_xas
from silx.io.url import DataUrl
from est.units import ur
import json
import pint
import silx.io.utils
from est.core.types import Dim

try:
    import PyMca5
except ImportError:
    has_pymca = False
else:
    has_pymca = True
    from PyMca5.PyMcaDataDir import PYMCA_DATA_DIR
    from est.io.utils.spec import read_spectrum


class TestSpectrum(unittest.TestCase):
    """Test the spectrum class"""

    @unittest.skipIf(has_pymca is False, "PyMca5 is not installed")
    def test_from_dat(self):
        """check that we can create a Spectrum from a pymca .dat file"""
        data_file = os.path.join(PYMCA_DATA_DIR, "EXAFS_Cu.dat")
        energy, mu = read_spectrum(
            data_file,
            energy_col_name="Column 1",
            absorption_col_name="Column 2",
        )
        spectrum = Spectrum(energy=energy, mu=mu)
        self.assertTrue(spectrum.energy is not None)
        self.assertTrue(spectrum.mu is not None)

    def test_from_numpy_array(self):
        """check that we can create a Spectrum from numpy arrays"""
        energy = numpy.arange(10, 20)
        mu = numpy.arange(10)
        spectrum = Spectrum(energy=energy, mu=mu)
        numpy.testing.assert_array_equal(spectrum.energy, (energy * ur.eV).m)
        numpy.testing.assert_array_equal(spectrum.mu, mu)
        mu_2 = numpy.arange(30, 40)
        spectrum["Mu"] = mu_2
        numpy.testing.assert_array_equal(spectrum.mu, mu_2)


class TestXASObject(unittest.TestCase):
    """test construction of XAS object from a single spectra"""

    @unittest.skipIf(has_pymca is False, "PyMca5 is not installed")
    def test_create_from_single_spectrum(self):
        """check that we can create a XASObject from a pymca .dat file"""
        data_file = os.path.join(PYMCA_DATA_DIR, "EXAFS_Cu.dat")
        spectrum = {}
        spectrum["Energy"], spectrum["Mu"] = read_spectrum(
            data_file,
            energy_col_name="Column 1",
            absorption_col_name="Column 2",
        )
        self.spectrum = Spectrum(energy=spectrum["Energy"], mu=spectrum["Mu"])
        self.configuration = {
            "FT": {"KWeight": 1},
            "EXAFS": {"EXAFSNormalized": numpy.array([1, 2, 3])},
        }
        obj = XASObject(
            spectra=(self.spectrum,),
            energy=self.spectrum.energy,
            configuration=self.configuration,
            dim1=1,
            dim2=1,
        )
        self.assertEqual(obj.n_spectrum, 1)
        ddict = obj.to_dict()
        obj2 = XASObject.from_dict(ddict)
        self.assertEqual(obj2, obj)
        # insure the XASObject is serializable
        # import json
        # json.dumps(obj2.to_dict())

    def test_create_from_several_spectrums(self):
        """check that we can create a XASObject from numpy arrays"""
        self.energy, self.spectra = spectra_utils.create_dataset(shape=(256, 20, 10))
        self.output_dir = tempfile.mkdtemp()
        spectra_path = "/data/NXdata/data"
        channel_path = "/data/NXdata/Channel"
        filename = os.path.join(self.output_dir, "myfile.h5")
        with h5py.File(filename, "a") as f:
            f[spectra_path] = self.spectra
            f[channel_path] = self.energy

        url_spectra = DataUrl(file_path=filename, data_path=spectra_path, scheme="silx")
        self.xas_obj = read_xas(
            spectra_url=url_spectra,
            channel_url=DataUrl(
                file_path=filename, data_path=channel_path, scheme="silx"
            ),
            dimensions=(Dim.DIM_2, Dim.DIM_1, Dim.DIM_0),
        )
        self.assertEqual(self.xas_obj.spectra.shape[0], 20)
        self.assertEqual(self.xas_obj.spectra.shape[1], 10)
        self.assertEqual(self.xas_obj.n_spectrum, 20 * 10)
        ddict = self.xas_obj.to_dict(with_process_details=False)
        original_spectra = silx.io.utils.get_data(
            DataUrl(file_path=filename, data_path=spectra_path, scheme="silx")
        )
        obj2 = XASObject.from_dict(ddict)
        self.assertEqual(self.xas_obj.n_spectrum, obj2.n_spectrum)
        obj2_mu_spectra = obj2.spectra.map_to(data_info="mu")

        numpy.testing.assert_array_equal(original_spectra, obj2_mu_spectra)
        self.assertEqual(obj2, self.xas_obj)


class TestXASObjectSerialization(unittest.TestCase):
    def setUp(self) -> None:
        self.energy, self.spectra = spectra_utils.create_dataset(shape=(256, 20, 10))
        self.output_dir = tempfile.mkdtemp()
        self.spectra_path = "/data/NXdata/data"
        self.channel_path = "/data/NXdata/Channel"
        self.filename = os.path.join(self.output_dir, "myfile.h5")
        with h5py.File(self.filename, "a") as f:
            f[self.spectra_path] = self.spectra
            f[self.channel_path] = self.energy

        self.dimensions = (Dim.DIM_2, Dim.DIM_1, Dim.DIM_0)

        self.url_spectra = DataUrl(
            file_path=self.filename, data_path=self.spectra_path, scheme="silx"
        )
        self.url_energy = DataUrl(
            file_path=self.filename, data_path=self.channel_path, scheme="silx"
        )
        self.process_flow_file = os.path.join(self.output_dir, "process_flow.h5")

    def tearDown(self) -> None:
        shutil.rmtree(self.output_dir)

    def test_serialization_url_auto(self):
        """Make sure the `to_dict` and `from_dict` functions are working
        if no url are provided"""
        xas_obj = XASObject(
            spectra=self.spectra,
            energy=self.energy,
            dim1=20,
            dim2=10,
            keep_process_flow=False,
        )
        # if no h5 file defined, should fail to copy it to a dictionary
        with self.assertRaises(ValueError):
            xas_obj.to_dict()

        xas_obj.link_to_h5(self.process_flow_file)
        dict_xas_obj = xas_obj.to_dict()

        # make sure it is serializable
        json.dumps(dict_xas_obj)
        # make sure we find a comparable xas object from it
        xas_obj_2 = XASObject.from_dict(dict_xas_obj)

        numpy.testing.assert_array_equal(xas_obj.energy, xas_obj_2.energy)
        self.assertEqual(xas_obj, xas_obj_2)

        # simple test without the process_details
        dict_xas_obj = xas_obj.to_dict(with_process_details=False)
        # make sure it is serializable
        json.dumps(dict_xas_obj)

    def test_serialization_url_provided(self):
        """Make sure the `to_dict` and `from_dict` functions are working
        if url are provided"""
        xas_obj = XASObject(
            spectra=self.spectra,
            energy=self.energy,
            dim1=20,
            dim2=10,
            keep_process_flow=False,
            energy_url=self.url_energy,
            spectra_url=self.spectra_path,
        )
        self.assertTrue(isinstance(xas_obj.energy, numpy.ndarray))
        # if no h5 file defined, should fail to copy it to a dictionary
        with self.assertRaises(ValueError):
            xas_obj.to_dict()

        xas_obj.link_to_h5(self.process_flow_file)
        dict_xas_obj = xas_obj.to_dict()

        # make sure it is serializable
        json.dumps(dict_xas_obj)
        # make sure we find a comparable xas object from it
        xas_obj_2 = XASObject.from_dict(dict_xas_obj)
        self.assertTrue(isinstance(xas_obj_2.energy, numpy.ndarray))

        numpy.testing.assert_array_equal(xas_obj.energy, xas_obj_2.energy)
        self.assertEqual(xas_obj, xas_obj_2)

        # simple test without the process_details
        dict_xas_obj = xas_obj.to_dict(with_process_details=False)
        # make sure it is serializable
        json.dumps(dict_xas_obj)


def suite():
    test_suite = unittest.TestSuite()
    for ui in (TestSpectrum, TestXASObject, TestXASObjectSerialization):
        test_suite.addTest(unittest.defaultTestLoader.loadTestsFromTestCase(ui))
    return test_suite


if __name__ == "__main__":
    unittest.main(defaultTest="suite")



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "07/30/2019"


from silx.gui import qt
from est.gui.larch.utils import _OptionalQDoubleSpinBox, _OptionalQIntSpinBox


class _MPreEdgeParameters(qt.QWidget):
    """
    Widget for setting the configuration of the larch 'pre edge' process
    """

    sigChanged = qt.Signal()
    """Signal emitted when parameters changed"""

    sigE0CalculatorRequest = qt.Signal()
    """Signal emitted when E0 calculator is requested"""

    def __init__(self, parent=None):
        qt.QWidget.__init__(self, parent)
        self.setLayout(qt.QGridLayout())

        # e0
        self._e0SB = _OptionalQDoubleSpinBox(parent=self)
        self._e0SB.setToolTip(
            "edge energy, in eV.  If None, it will be " "determined here."
        )
        self._e0SB.setMinimum(0.0)
        self._e0SB.setValue(2000.0)
        self._e0SB.setValue(None)
        self.layout().addWidget(qt.QLabel("e0", parent=self), 0, 0)
        self.layout().addWidget(self._e0SB, 0, 1)
        style = qt.QApplication.instance().style()
        e0_icon = style.standardIcon(qt.QStyle.SP_FileDialogContentsView)
        self._e0calculationPB = qt.QPushButton(e0_icon, "", parent=self)
        self._e0calculationPB.setToolTip(
            "Open E0 calculation dialog to compute" "mean or median E value from a roi"
        )

        self.layout().addWidget(self._e0calculationPB, 0, 2)
        # step
        self._stepSP = _OptionalQDoubleSpinBox(parent=self)
        self._stepSP.setToolTip("edge jump. If None, it will be determined")
        self._stepSP.setValue(None)
        self.layout().addWidget(qt.QLabel("step", parent=self), 1, 0)
        self.layout().addWidget(self._stepSP, 1, 1)
        # pre1
        self._pre1SB = _OptionalQDoubleSpinBox(parent=self)
        self._pre1SB.setToolTip("low E range (relative to e0) for pre-edge " "region")
        self._pre1SB.setValue(None)
        self._pre1SB.setContentsMargins(0, 0, 0, 0)
        self.layout().addWidget(qt.QLabel("pre1", parent=self), 2, 0)
        self.layout().addWidget(self._pre1SB, 2, 1)
        # pre2
        self._pre2SB = qt.QDoubleSpinBox(parent=self)
        self._pre2SB.setToolTip("high E range (relative to e0) for pre-edge " "region")
        self._pre2SB.setMinimum(-99999)
        self._pre2SB.setMaximum(99999)
        self._pre2SB.setValue(-50)
        self._pre2SB.setContentsMargins(0, 0, 0, 0)
        self.layout().addWidget(qt.QLabel("pre2", parent=self), 3, 0)
        self.layout().addWidget(self._pre2SB, 3, 1)
        # norm1
        self._norm1SB = qt.QDoubleSpinBox(parent=self)
        self._norm1SB.setToolTip("low E range (relative to e0) for post-edge " "region")
        self._norm1SB.setContentsMargins(0, 0, 0, 0)
        self._norm1SB.setMinimum(-99999)
        self._norm1SB.setMaximum(99999)
        self._norm1SB.setValue(100)
        self.layout().addWidget(qt.QLabel("norm1", parent=self), 4, 0)
        self.layout().addWidget(self._norm1SB, 4, 1)
        # norm2
        self._norm2SB = _OptionalQDoubleSpinBox(parent=self)
        self._norm2SB.setToolTip(
            "high E range (relative to e0) for post-edge " "region"
        )
        self._norm2SB.setValue(None)
        self._norm2SB.setContentsMargins(0, 0, 0, 0)
        self.layout().addWidget(qt.QLabel("norm2", parent=self), 5, 0)
        self.layout().addWidget(self._norm2SB, 5, 1)
        # nvict
        self._nvictSB = qt.QSpinBox(parent=self)
        self._nvictSB.setValue(0)
        self.layout().addWidget(qt.QLabel("energy exponent", parent=self), 6, 0)
        self.layout().addWidget(self._nvictSB, 6, 1)
        # nnorm
        self._nnorm = _OptionalQIntSpinBox(parent=self)
        self._nnorm.setRange(0, 3)
        self._nnorm.setValue(None)
        self._nnorm.setToolTip(
            "degree of polynomial (ie, nnorm+1 coefficients "
            "will be found) for post-edge normalization "
            "curve. If unset nnorm will default to 2 in "
            "norm2-norm1>400, to 1 if 100>norm2-norm1>300, "
            "and to 0 in norm2-norm1<100."
        )
        self.layout().addWidget(qt.QLabel("polynomial degree", parent=self), 7, 0)
        self.layout().addWidget(self._nnorm, 7, 1)
        # make_flat
        self._makeFlatCB = qt.QCheckBox("", parent=self)
        self._makeFlatCB.setToolTip("to calculate flattened output")
        self._makeFlatCB.setChecked(True)
        self.layout().addWidget(qt.QLabel("make flat", parent=self), 8, 0)
        self.layout().addWidget(self._makeFlatCB, 8, 1)
        # emin_area
        self._eminAreaSB = _OptionalQDoubleSpinBox(parent=self)
        self._eminAreaSB.setMinimum(0.0)
        self._eminAreaSB.setValue(None)
        self._eminAreaSB.setToolTip(
            "Energy threshold for area normalization."
            "norm_area will be estimated so that the "
            "area between emin_area and norm2 is equal "
            "to (norm2-emin_area). By default emin_area"
            " will be set to the *nominal* edge energy "
            "for the element and "
            "edge - 3*core_level_width"
        )
        self.layout().addWidget(qt.QLabel("e min area", parent=self), 9, 0)
        self.layout().addWidget(self._eminAreaSB, 9, 1)

        # expose API
        self.getStep = self._stepSP.getValue
        self.setStep = self._stepSP.setValue
        self.getPre1 = self._pre1SB.getValue
        self.setPre1 = self._pre1SB.setValue
        self.getPre2 = self._pre2SB.value
        self.setPre2 = self._pre2SB.setValue
        self.getNorm1 = self._norm1SB.value
        self.setNorm1 = self._norm1SB.setValue
        self.getNorm2 = self._norm2SB.getValue
        self.setNorm2 = self._norm2SB.setValue
        self.getNNorm = self._nnorm.getValue
        self.setNNorm = self._nnorm.setValue
        self.getNVict = self._nvictSB.value
        self.setNVict = self._nvictSB.setValue
        self.isUsingMakeFlat = self._makeFlatCB.isChecked
        self.setMakeFlat = self._makeFlatCB.setChecked
        self.getEMinArea = self._eminAreaSB.getValue
        self.setEMinArea = self._eminAreaSB.setValue

        # connect signal / Slot
        self._stepSP.sigChanged.connect(self._valueChanged)
        self._e0SB.sigChanged.connect(self._valueChanged)
        self._pre1SB.sigChanged.connect(self._valueChanged)
        self._pre2SB.editingFinished.connect(self._valueChanged)
        self._norm1SB.editingFinished.connect(self._valueChanged)
        self._norm2SB.sigChanged.connect(self._valueChanged)
        self._nnorm.sigChanged.connect(self._valueChanged)
        self._nvictSB.editingFinished.connect(self._valueChanged)
        self._makeFlatCB.toggled.connect(self._valueChanged)
        self._eminAreaSB.sigChanged.connect(self._valueChanged)
        self._e0calculationPB.released.connect(self._updateE0FromDialog)

        # avoid several emission of the sigChanged signal
        self._lastParameters = None

    def _valueChanged(self, *arg, **kwargs):
        current_values = self.getParameters()
        if current_values != self._lastParameters:
            self._lastParameters = current_values
            self.sigChanged.emit()

    def getParameters(self):
        return {
            "e0": self.getE0(),
            "step": self.getStep(),
            "pre1": self.getPre1(),
            "pre2": self.getPre2(),
            "norm1": self.getNorm1(),
            "norm2": self.getNorm2(),
            "nnorm": self.getNNorm(),
            "nvict": self.getNVict(),
            "make_flat": self.isUsingMakeFlat(),
            "emin_area": self.getEMinArea(),
        }

    def setParameters(self, parameters):
        assert isinstance(parameters, dict)
        for key, value in parameters.items():
            if key == "step":
                self.setStep(value)
            elif key == "e0":
                self.setE0(value)
            elif key == "pre1":
                self.setPre1(value)
            elif key == "pre2":
                self.setPre2(value)
            elif key == "norm1":
                self.setNorm1(value)
            elif key == "norm2":
                self.setNorm2(value)
            elif key == "nnorm":
                self.setNNorm(value)
            elif key == "nvict":
                self.setNVict(value)
            elif key == "make_flat":
                self.setMakeFlat(value)
            elif key == "emin_area":
                self.setEMinArea(value)

    def getE0(self):
        return self._e0SB.getValue()

    def setE0(self, value):
        self._e0SB.setValue(value=value)

    def _updateE0FromDialog(self):
        self.sigE0CalculatorRequest.emit()


if __name__ == "__main__":
    from est.gui.e0calculator import E0CalculatorDialog
    from est.core.utils import spectra as spectra_utils
    from est.core.types import XASObject
    import functools

    def create_dataset():
        dim1, dim2 = 20, 40

        energy, spectra = spectra_utils.create_dataset(
            shape=(256, dim1, dim2), noise=True
        )
        return XASObject(energy=energy, spectra=spectra, dim1=dim1, dim2=dim2)

    def launchE0Calculator(xas_object, pre_edge_widget):
        dialog = E0CalculatorDialog(xas_obj=xas_object, parent=None)
        if dialog.exec_():
            pre_edge_widget.setE0(dialog.getE0())

    xas_object = create_dataset()
    app = qt.QApplication([])
    widget = _MPreEdgeParameters()
    widget.sigE0CalculatorRequest.connect(
        functools.partial(launchE0Calculator, xas_object, widget)
    )

    widget.show()
    app.exec_()



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "07/30/2019"


from silx.gui import qt


class _OptionalQDoubleSpinBox(qt.QWidget):
    """
    Simple widget allowing to activate or tnoe the spin box
    """

    sigChanged = qt.Signal()
    """Signal emitted when parameters changed"""

    def __init__(self, parent):
        qt.QWidget.__init__(self, parent)
        self.setLayout(qt.QHBoxLayout())
        self._checkbox = qt.QCheckBox(parent=self)
        self.layout().addWidget(self._checkbox)
        self._spinBox = qt.QDoubleSpinBox(parent=self)
        self.layout().addWidget(self._spinBox)

        # default set up
        self._checkbox.setChecked(True)
        self._spinBox.setMinimum(-999999)
        self._spinBox.setMaximum(999999)
        self._lastValue = None

        # expose API
        self.setMinimum = self._spinBox.setMinimum
        self.setMaximum = self._spinBox.setMaximum

        # connect signal / slot
        self._checkbox.toggled.connect(self._updateSpinBoxStatus)
        self._spinBox.editingFinished.connect(self._valueChanged)

    def getValue(self):
        if self._checkbox.isChecked():
            return self._spinBox.value()
        else:
            return None

    def setValue(self, value):
        self._checkbox.setChecked(value is not None)
        if value is not None:
            self._spinBox.setValue(value)

    def _updateSpinBoxStatus(self, *arg, **kwargs):
        self._spinBox.setEnabled(self._checkbox.isChecked())
        self._valueChanged()

    def _valueChanged(self, *arg, **kwargs):
        if self._lastValue != self.getValue():
            self._lastValue = self.getValue()
            self.sigChanged.emit()


class _OptionalQIntSpinBox(qt.QWidget):
    """
    Simple widget allowing to activate or tnoe the spin box
    """

    sigChanged = qt.Signal()
    """Signal emitted when parameters changed"""

    def __init__(self, parent):
        qt.QWidget.__init__(self, parent)
        self.setLayout(qt.QHBoxLayout())
        self._checkbox = qt.QCheckBox(parent=self)
        self.layout().addWidget(self._checkbox)
        self._spinBox = qt.QSpinBox(parent=self)
        self.layout().addWidget(self._spinBox)

        # default set up
        self._checkbox.setChecked(True)
        self._lastValue = None

        # expose API
        self.setMinimum = self._spinBox.setMinimum
        self.setMaximum = self._spinBox.setMaximum
        self.setRange = self._spinBox.setRange

        # connect signal / slot
        self._checkbox.toggled.connect(self._updateSpinBoxStatus)
        self._spinBox.valueChanged.connect(self._valueChanged)

    def getValue(self):
        if self._checkbox.isChecked():
            return self._spinBox.value()
        else:
            return None

    def setValue(self, value):
        self._checkbox.setChecked(value is not None)
        if value is not None:
            self._spinBox.setValue(value)

    def _updateSpinBoxStatus(self, *arg, **kwargs):
        self._spinBox.setEnabled(self._checkbox.isChecked())
        self._valueChanged()

    def _valueChanged(self, *arg, **kwargs):
        if self._lastValue != self.getValue():
            self._lastValue = self.getValue()
            self.sigChanged.emit()



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "06/11/2019"


from .process import Process
from est.core.types import XASObject
from est.core.process.process import _input_desc
from est.core.process.process import _output_desc
import pkg_resources


class DumpXasObject(Process):

    inputs = [
        _input_desc(name="xas_obj", type=XASObject, handler="process", doc=""),
    ]

    outputs = [
        _output_desc(name="xas_obj", type=XASObject, doc=""),
    ]

    def __init__(self):
        super().__init__(name="dump xas object")
        self._output_file = None

    @property
    def output_file(self):
        return self._output_file

    @output_file.setter
    def output_file(self, output_file):
        self._output_file = output_file

    def program_name(self):
        """Name of the program used for this processing"""
        return "xas writer"

    def program_version(self):
        """version of the program used for this processing"""
        return pkg_resources.get_distribution("est").version

    def definition(self):
        """definition of the process"""
        return "write xas object to a file"

    def process(self, xas_object):
        if isinstance(xas_object, dict):
            xas_object = XASObject.from_dict(xas_object)
        if not isinstance(xas_object, XASObject):
            raise TypeError(
                "xas_object should be a convertable dict or an" "instance of XASObject"
            )

        if self.output_file is None:
            raise ValueError("output file not provided")
        else:
            xas_object.dump(self.output_file)

    def set_properties(self, properties):
        self.setConfiguration(properties)

    def setConfiguration(self, configuration):
        super().setConfiguration(configuration=configuration)
        self._read_output_file()

    def update_properties(self, properties):
        super().update_properties(properties=properties)
        self._read_output_file()

    def _read_output_file(self):
        if "output_file" in self._settings:
            self.output_file = self._settings["output_file"]
        elif "_output_file_setting" in self._settings:
            self.output_file = self._settings["_output_file_setting"]

    __call__ = process



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "14/12/2020"


import numpy
from typing import Union
from typing import Iterable
import logging
from est.units import units
from est.units import ur
from est.units import convert_to as convert_unit_to
import pint

_logger = logging.getLogger(__name__)
try:
    import larch
except ImportError:

    class _Spectrum_Base:
        pass

    _has_larch = False
    _logger.info("larch not found")
else:
    from larch.symboltable import Group as LarchGroup

    class _Spectrum_Base(LarchGroup):
        pass

    from est.core.utils import larchutils

    _has_larch = True
    _logger.info("larch found")


class Spectrum(_Spectrum_Base):
    """
    Core object to be used to store larch and pymca results.

    Larch is using 'Group' to store the results and adds members to this group
    according to the different treatment. Pymca is using a dictionary to store
    the results.

    This class has to adpat to both behaviors and the different naming
    convention as well.

    :param numpy.ndarray (1D) energy: beam energy
    :param numpy.ndarray (1D) mu: beam absorption
    :param int x: x index on the spectra
    :param int y: y index on the spectra
    """

    _MU_KEY = "Mu"

    _ENERGY_KEY = "Energy"

    _NORMALIZED_MU_KEY = "NormalizedMu"

    _NORMALIZED_ENERGY_KEY = "NormalizedEnergy"

    _NORMALIZED_SIGNAL_KEY = "NormalizedSignal"

    _EXAFS_SIGNAL_KEY = "EXAFSSignal"

    _EXAFS_KVALUES_KEY = "EXAFSKValues"

    _FT_KEY = "FT"

    _EDGE_KEY = "Edge"

    _NORMALIZED_BACKGROUND_KEY = "NormalizedBackground"

    _X_POS_KEY = "XPos"

    _Y_POS_KEY = "YPos"

    def __init__(
        self,
        energy: Union[None, numpy.ndarray] = None,
        mu: Union[None, numpy.ndarray] = None,
        x: Union[None, int] = None,
        y: Union[None, int] = None,
    ):
        super().__init__()
        if energy is not None:
            assert isinstance(energy, (numpy.ndarray, pint.Quantity))
            if isinstance(energy, numpy.ndarray):
                energy = energy * ur.eV

        self.__x = x
        self.__y = y

        # properties
        self.energy = energy
        self.mu = mu
        self.__chi = None
        self.__k_values = None
        self.__normalized_mu = None
        self.__normalized_energy = None
        self.__pre_edge = None
        self.__post_edge = None
        # unable to create a property top level since larch is using
        # getattr(group, 'edge', None). Or edge should be initialized to
        # something != None that we don't want
        self.__e0 = None
        self.__noise_savgol = None
        self.__norm_noise_savgol = None
        self.__other_parameters = {}
        self.ft = {}

        self.__key_mapper = {
            self._MU_KEY: self.__class__.mu,
            self._ENERGY_KEY: self.__class__.energy,
            self._NORMALIZED_MU_KEY: self.__class__.normalized_mu,
            self._NORMALIZED_ENERGY_KEY: self.__class__.normalized_energy,
            self._NORMALIZED_SIGNAL_KEY: self.__class__.post_edge,
            self._NORMALIZED_BACKGROUND_KEY: self.__class__.pre_edge,
            self._FT_KEY: self.__class__.ft,
            self._EDGE_KEY: self.__class__.e0,
            self._EXAFS_SIGNAL_KEY: self.__class__.chi,
            self._EXAFS_KVALUES_KEY: self.__class__.k,
        }

    @property
    def energy(self) -> Union[None, numpy.ndarray]:
        """

        :note: cannot be with unit because use directly by xraylarch and pymca
        """
        return self.__energy

    @energy.setter
    @units(energy="eV")
    def energy(self, energy):
        if isinstance(energy, pint.Quantity):
            self.__energy = energy.m_as(ur.eV)
        else:
            self.__energy = energy

    @property
    def mu(self) -> Union[None, numpy.ndarray]:
        return self.__mu

    @mu.setter
    def mu(self, mu: numpy.ndarray):
        assert isinstance(mu, numpy.ndarray) or mu is None
        self.__mu = mu

    @property
    def x(self) -> Union[None, int]:
        return self.__x

    @property
    def y(self) -> Union[None, int]:
        return self.__y

    @property
    def chi(self) -> Union[None, numpy.ndarray]:
        return self.__chi

    @chi.setter
    def chi(self, chi: numpy.ndarray):
        self.__chi = chi

    @property
    def k(self) -> Union[None, numpy.ndarray]:
        return self.__k_values

    @k.setter
    def k(self, k: numpy.ndarray):
        self.__k_values = k

    @property
    def normalized_mu(self) -> Union[None, numpy.ndarray]:
        return self.__normalized_mu

    @normalized_mu.setter
    def normalized_mu(self, mu: numpy.ndarray):
        assert isinstance(mu, numpy.ndarray) or mu is None
        self.__normalized_mu = mu

    @property
    def norm(self) -> Union[None, numpy.ndarray]:
        # this alias is needed for larch
        return self.__normalized_mu

    @norm.setter
    def norm(self, value: numpy.ndarray):
        # this alias is needed for larch
        self.__normalized_mu = value

    @property
    def normalized_energy(self) -> Union[None, numpy.ndarray]:
        return self.__normalized_energy

    @normalized_energy.setter
    def normalized_energy(self, energy: Union[numpy.ndarray, pint.Quantity]):
        assert isinstance(energy, (numpy.ndarray, pint.Quantity)) or energy is None
        self.__normalized_energy = energy

    @property
    def pre_edge(self) -> Union[None, numpy.ndarray]:
        return self.__pre_edge

    @pre_edge.setter
    def pre_edge(self, value: numpy.ndarray):
        self.__pre_edge = value

    @property
    def _edge(self) -> Union[None, numpy.ndarray]:
        return self.edge

    @_edge.setter
    def _edge(self, value: numpy.ndarray):
        self.edge = value

    @property
    def post_edge(self) -> Union[None, numpy.ndarray]:
        return self.__post_edge

    @post_edge.setter
    def post_edge(self, value: numpy.ndarray):
        self.__post_edge = value

    @property
    def e0(self) -> Union[None, numpy.ndarray]:
        return self.__e0

    @e0.setter
    def e0(self, e0: numpy.ndarray):
        self.__e0 = e0

    @property
    def noise_savgol(self):
        return self.__noise_savgol

    @noise_savgol.setter
    def noise_savgol(self, values):
        self.__noise_savgol = values

    @property
    def norm_noise_savgol(self):
        return self.__norm_noise_savgol

    @norm_noise_savgol.setter
    def norm_noise_savgol(self, values):
        self.__norm_noise_savgol = values

    @property
    def ft(self):
        return self.__ft

    @ft.setter
    def ft(self, ft):
        if isinstance(ft, _FT):
            self.__ft = ft
        else:
            self.__ft = _FT(ddict=ft)

    @property
    def shape(self) -> tuple:
        _energy_len = 0
        if self.__energy is not None:
            _energy_len = len(self.__energy)
        _mu_len = 0
        if self.__mu is not None:
            _mu_len = len(self.__mu)

        return (_energy_len, _mu_len)

    def extra_keys(self) -> tuple:
        return tuple(self.__other_parameters.keys())

    def __getitem__(self, key):
        """Need for pymca compatibility"""
        if key in self.__key_mapper:
            return self.__key_mapper[key].fget(self)
        else:
            if _has_larch and key in self._members():
                return self._members()[key]
            else:
                return self.__other_parameters[key]

    def __setitem__(self, key, value):
        """Need for pymca compatibility"""
        assert key
        if key in self.__key_mapper:
            self.__key_mapper[key].fset(self, value)
        else:
            self.__other_parameters[key] = value

    def __contains__(self, item):
        return item in self.__key_mapper or item in self.__other_parameters

    def load_frm_dict(self, ddict: dict):
        assert isinstance(ddict, dict)

        def value_is_none(value):
            if hasattr(value, "decode"):
                value = value.decode("UTF-8")

            if isinstance(value, str):
                return value == "None"
            else:
                return value is None

        larch_ddict = {}
        if "_larch_grp_members" in ddict:
            if _has_larch is False:
                _logger.warning(
                    "larch is not installed but the Spectrum "
                    "requires it, won't be able to load data "
                    "relative to larch"
                )
            for key in ddict["_larch_grp_members"]:
                if hasattr(key, "decode"):
                    key = key.decode("UTF-8")
                larch_ddict[key] = None if value_is_none(ddict[key]) else ddict[key]
                del ddict[key]

        for key, value in ddict.items():
            if hasattr(value, "decode"):
                value = value.decode("UTF-8")
            self[key] = None if value_is_none(value) else value
        for key, value in larch_ddict.items():
            setattr(self, key, None if value_is_none(value) else value)
        return self

    @staticmethod
    def from_dict(ddict: dict):
        x_pos = None
        y_pos = None
        if Spectrum._X_POS_KEY in ddict:
            x_pos = ddict[Spectrum._X_POS_KEY]
        if Spectrum._Y_POS_KEY in ddict:
            y_pos = ddict[Spectrum._Y_POS_KEY]
        spectrum = Spectrum(x=x_pos, y=y_pos)
        return spectrum.load_frm_dict(ddict=ddict)

    def to_dict(self) -> dict:
        energy = convert_unit_to(self.energy, ur.eV)
        if energy is not None:
            energy = energy.m_as(ur.eV)
        res = {
            self._X_POS_KEY: self.x,
            self._Y_POS_KEY: self.y,
            self._MU_KEY: self.mu,
            self._ENERGY_KEY: energy,
            self._FT_KEY: self.ft.to_dict(),
            self._NORMALIZED_MU_KEY: "None"
            if self.normalized_mu is None
            else self.normalized_mu,
            self._NORMALIZED_ENERGY_KEY: "None"
            if self.normalized_energy is None
            else self.normalized_energy,
            self._NORMALIZED_SIGNAL_KEY: "None"
            if self.post_edge is None
            else self.post_edge,
            self._NORMALIZED_BACKGROUND_KEY: "None"
            if self.pre_edge is None
            else self.pre_edge,
            self._EDGE_KEY: "None" if self.e0 is None else self.e0,
            self._EXAFS_SIGNAL_KEY: "None" if self.chi is None else self.chi,
            self._EXAFS_KVALUES_KEY: "None" if self.k is None else self.k,
        }
        if _has_larch:
            res.update(self._getLarchGroupMenbers())
        res.update(self.__other_parameters)
        return res

    def _getLarchGroupMenbers(self):
        """Return larch group specific menbers"""

        assert _has_larch is True
        res = {}
        res["_larch_grp_members"] = list(self._members().keys())
        for key in self._members().keys():
            if isinstance(self[key], LarchGroup):
                res[key] = larchutils.group_to_dict(self[key])
            else:
                res[key] = getattr(self, key)
        return res

    def __str__(self):
        def add_info(str_, attr):
            assert hasattr(self, attr)
            sub_str = "- " + attr + ": " + str(getattr(self, attr)) + "\n"
            return str_ + sub_str

        main_info = ""
        for info in (
            "energy",
            "mu",
            "normalized_mu",
            "normalized_signal",
            "normalized_energy",
        ):
            main_info = add_info(str_=main_info, attr=info)

        def add_third_info(str_, key):
            sub_str = ("- " + key + ": " + str(self[key])) + "\n"
            return str_ + sub_str

        for key in self.__other_parameters:
            main_info = add_third_info(str_=main_info, key=key)
        return main_info

    def update(self, obj):
        """
        Update the contained values from the given obj.

        :param obj:
        :type obj: Union[XASObject, dict]
        """
        if isinstance(obj, Spectrum):
            _obj = obj.to_dict()
        else:
            _obj = obj
        assert isinstance(_obj, dict)
        for key, value in _obj.items():
            if isinstance(value, str) and value == "None":
                self[key] = None
            else:
                self[key] = value

    def get_missing_keys(self, keys: Iterable) -> tuple:
        """Return missing keys on the spectrum"""
        missing = []
        for key in keys:
            if key not in self or self[key] is None:
                missing.append(key)
        return tuple(missing)

    def keys(self) -> list:
        keys = list(self.__other_parameters.keys())
        keys += list(self.__key_mapper.keys())
        return keys

    def copy(self):
        return Spectrum().load_frm_dict(self.to_dict())

    def _force_indexes(self, x, y):
        """This is protected because it might change display and
        the indexes should be defined during Spectra or Spectrum construction
        once for all"""
        self.__x = x
        self.__y = y

    # def formula_results(self, formula):


class _FT(object):

    _RADIUS_KEY = "FTRadius"

    _INTENSITY_KEY = "FTIntensity"

    _IMAGINARY_KEY = "FTImaginary"

    def __init__(self, ddict):
        self.__radius = numpy.nan
        self.__intensity = numpy.nan
        self.__imaginary = numpy.nan
        self.__other_parameters = {}

        self.__key_mapper = {
            self._RADIUS_KEY: self.__class__.radius,
            self._INTENSITY_KEY: self.__class__.intensity,
            self._IMAGINARY_KEY: self.__class__.imaginary,
        }

        if ddict is not None:
            for key, values in ddict.items():
                self[key] = values

    @property
    def radius(self):
        return self.__radius

    @radius.setter
    def radius(self, radius):
        self.__radius = radius

    @property
    def intensity(self):
        return self.__intensity

    @intensity.setter
    def intensity(self, intensity):
        self.__intensity = intensity

    @property
    def imaginary(self):
        return self.__imaginary

    @imaginary.setter
    def imaginary(self, imaginery):
        self.__imaginary = imaginery

    def __getitem__(self, key):
        """Need for pymca compatibility"""
        if key in self.__key_mapper:
            return self.__key_mapper[key].fget(self)
        else:
            return self.__other_parameters[key]

    def __setitem__(self, key, value):
        """Need for pymca compatibility"""
        if key in self.__key_mapper:
            self.__key_mapper[key].fset(self, value)
        else:
            self.__other_parameters[key] = value

    def __contains__(self, item):
        return item in self.__key_mapper or item in self.__other_parameters

    def to_dict(self) -> dict:
        res = {
            self._RADIUS_KEY: self.radius,
            self._INTENSITY_KEY: self.intensity,
            self._IMAGINARY_KEY: self.imaginary,
        }
        res.update(self.__other_parameters)
        return res

    def get_missing_keys(self, keys: Iterable) -> tuple:
        """Return missing keys on the spectrum"""
        missing = []
        for key in keys:
            if key not in self:
                missing.append(key)
        return tuple(missing)



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "10/12/2020"


from silx.io.url import DataUrl
from est.units import ur
from est.core.types import Dim
from est.thirdparty.silx.hdf5file import get_data
from est.core.types import Spectrum, XASObject
from est.core.io import read_xas
from est.io.utils.information import InputInformation
import logging

try:
    from est.io.utils.spec import read_spectrum

    has_read_spectrum = True
except ImportError:
    has_read_spectrum = False
_logger = logging.getLogger(__name__)


def convert_spectra_dims(dims):
    """
    Convert a tuple of dims that can be strings... to a tuple of
    est.core.types.Dim
    """
    if dims is None:
        return None
    res = []
    for dim in dims:
        if isinstance(dim, str):
            if dim.lower() == Dim.DIM_0.value.lower():
                dim = Dim.DIM_0
            elif dim.lower() == Dim.DIM_1.value.lower():
                dim = Dim.DIM_1
            elif dim.lower() == Dim.DIM_2.value.lower():
                dim = Dim.DIM_2

        dim = Dim.from_value(dim)
        if dim in res:
            raise ValueError(
                "dimension {} has been provided several time."
                "Each dimension should be set once.".format(dim.value())
            )
        else:
            res.append(dim)
    return tuple(dims)


def get_unit(current_unit):
    if current_unit == "eV":
        return ur.eV
    elif current_unit == "keV":
        return ur.keV
    elif current_unit == "J":
        return ur.J
    elif current_unit == "kJ":
        return ur.kJ
    elif isinstance(current_unit, ur.Quantity):
        return current_unit
    else:
        raise ValueError("{} is not a valid unit for quantity".format(current_unit))


def get_url(my_str):
    if my_str in (None, ""):
        return None
    else:
        assert isinstance(my_str, str)
        if "@" in my_str:
            try:
                entry, file_path = my_str.split("@")
            except Exception:
                pass
            else:
                return DataUrl(file_path=file_path, data_path=entry, scheme="silx")
        else:
            try:
                url = DataUrl(path=my_str)
            except Exception:
                pass
            else:
                return url
        raise ValueError("unrecognized url {}".format(my_str))


def get_xas_obj(input_information: InputInformation):
    """load xas object from command line input
    :param spec_input: tuple of Union[str, None, dict]
    """
    print("#### call get_xas_obj")
    if input_information.is_spec_input():
        if has_read_spectrum:
            energy, mu = read_spectrum(
                spec_file=input_information.spec_file,
                energy_col_name=input_information.input_energy_col_name,
                absorption_col_name=input_information.input_abs_col_name,
                monitor_col_name=input_information.input_monitor_col_name,
                energy_unit=input_information.energy_unit,
                scan_header_S=input_information.scan_title,
            )
            spectrum = Spectrum(energy=energy, mu=mu)
            xas_obj = XASObject(energy=energy, spectra=(spectrum,), dim1=1, dim2=1)
        else:
            raise ValueError("Unable to read spectrum")
    else:
        sp, en, conf = read_xas(information=input_information)
        xas_obj = XASObject(spectra=sp, energy=en, configuration=conf)
    attr_names = "I0", "I1", "I2", "mu_ref"
    attr_values = (
        getattr(input_information, "I0", None),
        getattr(input_information, "I1", None),
        getattr(input_information, "I2", None),
        getattr(input_information, "mu_ref", None),
    )
    for attr_name, attr_value in zip(attr_names, attr_values):
        if attr_value is not None:
            if isinstance(attr_value, DataUrl):
                data = get_data(attr_value)
            else:
                data = attr_value
            if data.shape[0] > xas_obj.energy.shape[0]:
                _logger.warning(
                    "energy has less value than dataset. Clip {}.".format(attr_name)
                )
                data = data[: xas_obj.energy.shape[0]]
            xas_obj.attach_3d_array(attr_name, data)
    return xas_obj



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "18/01/2021"

import logging

from Orange.widgets import gui
from Orange.widgets.settings import Setting
from Orange.widgets.widget import OWWidget
from Orange.widgets.widget import Input, Output
from orangecontrib.est.progress import QProgress
from orangecontrib.est.process import ProcessRunnable
import Orange.data
import functools
from orangecontrib.est.utils import Converter
from orangecontrib.est.process import _ProcessForOrangeMixIn
from silx.gui import qt
import est.core.process.noise
from est.core.types import XASObject
from est.gui.noise import SavitskyGolayNoise

_logger = logging.getLogger(__file__)


class NoiseOW(_ProcessForOrangeMixIn, OWWidget):
    """
    Widget used to make the selection of a region of Interest to treat in a
    Dataset.
    """

    name = "noise"
    id = "orange.widgets.xas.utils.noise"
    description = "Compute noise using Savitsky-Golay"
    icon = "icons/noise.svg"

    priority = 40
    category = "esrfWidgets"
    keywords = ["dataset", "data", "noise"]

    process_function = est.core.process.noise.NoiseProcess

    want_main_area = True
    resizing_enabled = True
    compress_signal = False
    allows_cycle = False

    window_size = Setting(int(-1))
    polynomial_order = Setting(int(-1))
    e_min = Setting(int(-1))
    e_max = Setting(int(-1))

    class Inputs:
        xas_obj = Input("xas_obj", XASObject, default=True)
        # simple compatibility for some Orange widget and especialy the
        # 'spectroscopy add-on'
        data_table = Input("Data", Orange.data.Table)

    class Outputs:
        res_xas_obj = Output("xas_obj", XASObject)
        # by default we want to avoid sending 'Orange.data.Table' to avoid
        # loosing the XASObject flow process and results.

    def __init__(self):
        super().__init__()
        self._latest_xas_obj = None

        self._window = SavitskyGolayNoise(parent=self)
        layout = gui.vBox(self.mainArea, "noise").layout()
        layout.addWidget(self._window)

        # buttons
        types = qt.QDialogButtonBox.Ok
        self._buttons = qt.QDialogButtonBox(parent=self)
        self._buttons.setStandardButtons(types)
        layout.addWidget(self._buttons)

        self._buttons.hide()

        # expose API
        self.setWindowSize = self._window.setWindowSize
        self.getWindowSize = self._window.getWindowSize
        self.setPolynomialOrder = self._window.setPolynomialOrder
        self.getPolynomialOrder = self._window.getPolynomialOrder

        # manage settings
        self._window.setParameters(
            {
                "window_size": self.window_size,
                "polynomial_order": self.polynomial_order,
                "e_min": self.e_min,
                "e_max": self.e_max,
            }
        )

        # signal / slot connection
        # connect signals / slots
        self._window.sigChanged.connect(self._updateProcess)

    def _updateProcess(self, *arv, **kwargs):
        self._update_settings()
        if self._latest_xas_obj:
            self.process(xas_obj=self._latest_xas_obj)

    @Inputs.data_table
    def processFrmDataTable(self, data_table):
        if data_table is None:
            return
        self.process(Converter.toXASObject(data_table=data_table))

    @Inputs.xas_obj
    def process(self, xas_obj):
        if xas_obj is None:
            return

        if not self._canProcess():
            _logger.warning(
                "There is some processing on going already, will"
                "not process the new dataset"
            )

        self._latest_xas_obj = xas_obj
        self._startProcess()

        # setup the ft process
        process_obj = QNoise()
        process_obj._advancement.sigProgress.connect(self._setProgressValue)
        process_obj.set_properties({"noise": self._window.getParameters()})

        # update the processing thread
        thread = self.getProcessingThread()
        thread.init(process_obj=process_obj, xas_obj=self._latest_xas_obj)
        self._callback_finish = functools.partial(
            self._endProcess, self._latest_xas_obj
        )
        thread.finished.connect(self._callback_finish)

        # start processing
        thread.start(priority=qt.QThread.LowPriority)

    def _update_settings(self):
        self.window_size = self._window.getWindowSize()
        self.polynomial_order = self._window.getPolynomialOrder()
        self.e_min = self._window.getEMin()
        self.e_max = self._window.getEMax()

    def _setProgressValue(self, value):
        self._progress.widget.progressBarSet(value)


class QNoise(est.core.process.noise.NoiseProcess):
    """
    Normalization able to give advancement using qt.Signal and QThreadPool
    """

    def __init__(self):
        est.core.process.noise.NoiseProcess.__init__(self)
        self._advancement = QProgress("noise")

    def _pool_process(self, xas_obj):
        self.pool = qt.QThreadPool()
        self.pool.setMaxThreadCount(5)
        for spectrum in xas_obj.spectra:
            runnable = ProcessRunnable(
                fct=est.core.process.noise.process_noise_savgol,
                spectrum=spectrum,
                configuration=xas_obj.configuration,
                callback=self._advancement.increaseAdvancement,
            )
            self.pool.start(runnable)
        self.pool.waitForDone()



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2017-2019 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "06/11/2019"


import os
import shutil
import tempfile
import unittest
import h5py
import numpy
from silx.io.url import DataUrl
from est.core.io import read as read_xas
from est.core.process.roi import xas_roi
from est.core.utils import spectra as spectra_utils
from est.core.types import Dim


class TestRoi(unittest.TestCase):
    def setUp(self):
        self.energy, self.spectra = spectra_utils.create_dataset(shape=(16, 100, 30))
        self.output_dir = tempfile.mkdtemp()
        spectra_path = "/data/NXdata/data"
        channel_path = "/data/NXdata/Channel"
        filename = os.path.join(self.output_dir, "myfile.h5")
        with h5py.File(filename, "a") as f:
            f[spectra_path] = self.spectra
            f[channel_path] = self.energy

        self.xas_obj = read_xas(
            spectra_url=DataUrl(
                file_path=filename, data_path=spectra_path, scheme="silx"
            ),
            channel_url=DataUrl(
                file_path=filename, data_path=channel_path, scheme="silx"
            ),
            dimensions=(Dim.DIM_2, Dim.DIM_1, Dim.DIM_0),
        )

    def tearDown(self):
        shutil.rmtree(path=self.output_dir)

    def testApplyRoi(self):
        """Test output of the roi process"""
        original_spectra = self.xas_obj.spectra.map_to(data_info="mu").copy()
        self.assertEqual(original_spectra.shape, (16, 100, 30))
        roi_dict = {"origin": (20, 50), "size": (10, 20)}
        self.xas_obj.configuration = {"roi": roi_dict}
        res_xas_obj = xas_roi(self.xas_obj)
        self.assertEqual(res_xas_obj.n_spectrum, 20 * 10)
        reduces_spectra = res_xas_obj.spectra.map_to(data_info="mu").copy()
        assert reduces_spectra.shape == (16, 20, 10)
        numpy.testing.assert_array_equal(
            original_spectra[:, 50:70, 20:30], reduces_spectra
        )

    def testWorkflowWithRoi(self):
        """"""
        pass


def suite():
    test_suite = unittest.TestSuite()
    for ui in (TestRoi,):
        test_suite.addTest(unittest.defaultTestLoader.loadTestsFromTestCase(ui))
    return test_suite


if __name__ == "__main__":
    unittest.main(defaultTest="suite")



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
"""wrapper to the larch mback process"""

from est.core.types import Spectrum, XASObject
from est.core.process.process import Process
from est.core.process.process import _NexusSpectrumDef
from est.core.process.process import _NexusDatasetDef
from est.core.process.process import _input_desc
from est.core.process.process import _output_desc
from larch.xafs.xafsft import xftf
from est.core.utils.symbol import ANGSTROM_CHAR
import multiprocessing
import functools
import logging
import numpy

_logger = logging.getLogger(__name__)

_DEBUG = True
if _DEBUG:
    from larch.symboltable import Group


def process_spectr_xftf(
    spectrum,
    configuration,
    overwrite=True,
    callbacks=None,
    output=None,
    output_dict=None,
):
    """

    :param spectrum: spectrum to process
    :type: :class:`.Spectrum`
    :param configuration: configuration of the pymca normalization
    :type: dict
    :param overwrite: False if we want to return a new Spectrum instance
    :type: bool
    :param callback: callback to execute.
    :param output: list to store the result, needed for pool processing
    :type: multiprocessing.manager.list
    :param output_dict: key is input spectrum, value is index in the output
                        list.
    :type: dict
    :return: processed spectrum
    :rtype: tuple (configuration, spectrum)
    """
    _logger.debug("start xftf on spectrum (%s, %s)" % (spectrum.x, spectrum.y))
    assert isinstance(spectrum, Spectrum)
    if (
        not hasattr(spectrum, "k")
        or not hasattr(spectrum, "chi")
        or spectrum.k is None
        or spectrum.chi is None
    ):
        _logger.error(
            "k and/or chi is/are not specified, unable to compute "
            "xftf. Maybe you need to run autobk process before ?"
        )
        return None, None
    _conf = configuration
    if "xftf" in _conf:
        _conf = _conf["xftf"]
    opts = {}
    for opt_name in (
        "kmin",
        "kmax",
        "kweight",
        "dk",
        "dk2",
        "with_phase",
        "window",
        "rmax_out",
        "nfft",
        "kstep",
    ):
        if opt_name in _conf:
            opts[opt_name] = _conf[opt_name]
            if opt_name == "kweight":
                opts["kw"] = _conf[opt_name]
    if _DEBUG is True:
        assert isinstance(spectrum, Group)
    if overwrite:
        _spectrum = spectrum
    else:
        _spectrum = Spectrum().load_frm_dict(spectrum.to_dict())
    xftf(_spectrum, **opts)
    # handle chi(x) * k**k_weight plot with r max
    if hasattr(_spectrum, "k") and hasattr(_spectrum, "chi"):
        if "kweight" in opts:
            kweight = opts["kweight"]
        else:
            kweight = 0
        if (
            "_plot_chi_k_weighted_r_max" in _conf
            and _conf["_plot_chi_k_weighted_r_max"] is not None
        ):
            mask = _spectrum.k < _conf["_plot_chi_k_weighted_r_max"]
        else:
            mask = numpy.ones_like(_spectrum.k).astype(numpy.bool)
    _spectrum.masked_chi_weighted_k = _spectrum.chi[mask] * (
        _spectrum.k[mask] ** kweight
    )
    _spectrum.masked_k = _spectrum.k[mask]
    # handle ft plot with r max
    if hasattr(_spectrum, "r") and hasattr(_spectrum, "chir_mag"):
        if "_plot_chi_mag_r_max" in _conf and _conf["_plot_chi_mag_r_max"] is not None:
            mask = _spectrum.r < _conf["_plot_chi_mag_r_max"]
        else:
            mask = numpy.ones_like(_spectrum.r).astype(numpy.bool)
        _spectrum.masked_chir_mag = _spectrum.chir_mag[mask]
        _spectrum.masked_r = _spectrum.r[mask]

    if callbacks:
        for callback in callbacks:
            callback()
    return configuration, _spectrum


def larch_xftf(xas_obj):
    """

    :param xas_obj: object containing the configuration and spectra to process
    :type: Union[XASObject, dict]
    :return: spectra dict
    :rtype: XASObject
    """
    mback_obj = Larch_xftf()
    return mback_obj.process(xas_obj=xas_obj)


_USE_MULTIPROCESSING_POOL = False
# note: we cannot use multiprocessing pool with pypushflow for now.


class Larch_xftf(Process):

    inputs = [
        _input_desc(name="xas_obj", type=XASObject, handler="process", doc=""),
    ]

    outputs = [
        _output_desc(name="xas_obj", type=XASObject, doc=""),
    ]

    def __init__(self):
        Process.__init__(self, name="xftf")
        self._plot_settings = {}

    def set_properties(self, properties):
        if "_larchSettings" in properties:
            self._settings = properties["_larchSettings"]
        if "plot_settings" in properties:
            self._plot_settings = properties["plot_settings"]

    def process(self, xas_obj):
        _xas_obj = self.getXasObject(xas_obj=xas_obj)
        if self._settings:
            _xas_obj.configuration["xftf"] = self._settings
        elif "xftf" not in _xas_obj.configuration:
            _xas_obj.configuration["xftf"] = {}

        self._advancement.reset(max_=_xas_obj.n_spectrum)
        self._advancement.startProcess()
        self._pool_process(xas_obj=_xas_obj)
        self._advancement.endProcess()
        self.register_process(
            _xas_obj,
            data_keys=(
                _NexusDatasetDef(
                    "chir_im",
                    units="{}^(-3)".format(ANGSTROM_CHAR),
                    units_latex="\mathring{A}^{-3}",
                ),
                _NexusDatasetDef("chir_re"),
                _NexusDatasetDef(
                    "chir_mag",
                    units="{}^(-3)".format(ANGSTROM_CHAR),
                    units_latex="\mathring{A}^{-3}",
                ),
                _NexusDatasetDef(
                    "masked_chir_mag",
                    units="{}".format(ANGSTROM_CHAR),
                    units_latex="\mathring{A}",
                ),
                _NexusDatasetDef(
                    "r", units="{}".format(ANGSTROM_CHAR), units_latex="\mathring{A}"
                ),
                _NexusDatasetDef(
                    "masked_r",
                    units="{}".format(ANGSTROM_CHAR),
                    units_latex="\mathring{A}",
                ),
                _NexusDatasetDef(
                    "k",
                    units="{}^(-1)".format(ANGSTROM_CHAR),
                    units_latex="\mathring{A}^{-1}",
                ),
                _NexusDatasetDef(
                    "masked_k",
                    units="{}^(-1)".format(ANGSTROM_CHAR),
                    units_latex="\mathring{A}^{-1}",
                ),
                _NexusDatasetDef(
                    "masked_chi_weighted_k",
                    units="{}^(-2)".format(ANGSTROM_CHAR),
                    units_latex="\mathring{A}^{-2}",
                ),
            ),
            plots=(
                _NexusSpectrumDef(
                    signal="masked_chi_weighted_k",
                    axes=("masked_k",),
                    auxiliary_signals=None,
                    silx_style={"signal_scale_type": "linear"},
                    title="chi(k)*k^k_weight",
                    title_latex="\chi(k).k^{k\_weight}",
                ),
                _NexusSpectrumDef(
                    signal="masked_chir_mag",
                    axes=("masked_r",),
                    auxiliary_signals=None,
                    silx_style={"signal_scale_type": "linear"},
                    title="chir_mag",
                ),
            ),
        )
        return _xas_obj

    def _pool_process(self, xas_obj):
        assert isinstance(xas_obj, XASObject)
        if not _USE_MULTIPROCESSING_POOL:
            for spectrum in xas_obj.spectra:
                process_spectr_xftf(
                    spectrum=spectrum,
                    configuration=xas_obj.configuration,
                    callbacks=self.callbacks,
                    overwrite=True,
                )
        else:
            from multiprocessing import Manager

            manager = Manager()
            output_dict = {}
            res_list = manager.list()
            for i_spect, spect in enumerate(xas_obj.spectra):
                res_list.append(None)
                output_dict[spect] = i_spect

            with multiprocessing.Pool(5) as p:
                partial_ = functools.partial(
                    process_spectr_xftf,
                    configuration=xas_obj.configuration,
                    callbacks=self.callbacks,
                    overwrite=False,
                    output=res_list,
                    output_dict=output_dict,
                )
                p.map(partial_, xas_obj.spectra)

            # then update local spectrum
            for spectrum, res in zip(xas_obj.spectra, res_list):
                spectrum.update(res)

    def definition(self):
        return "xftf calculation"

    def program_version(self):
        import larch.version

        return larch.version.version_data()["larch"]

    def program_name(self):
        return "larch_xftf"

    __call__ = process



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

__authors__ = ["H. Payno", "Orange"]
__license__ = "MIT"
__date__ = "03/03/2017"

"""
Unit test from the Orange canvas main instance

"""
import os
import sys
import gc
import re
import logging
import pickle
import pkg_resources
from silx.gui.qt import QApplication
from AnyQt.QtGui import QFont
from AnyQt.QtCore import Qt, QDir
import AnyQt.importhooks

if AnyQt.USED_API == "pyqt5":
    # Use a backport shim to fake leftover PyQt4 imports
    AnyQt.importhooks.install_backport_hook("pyqt4")

from Orange import canvas
from est.gui.qtapplicationmanager import QApplicationManager
from Orange.canvas.application.canvasmain import CanvasMainWindow
from Orange.canvas.application.outputview import TextStream
from Orange.canvas.config import cache_dir
from Orange.canvas import config
from Orange.canvas.registry.base import WidgetRegistry
from Orange.canvas.registry import qt
from Orange.canvas.registry import set_global_registry
from Orange.canvas.scheme import SchemeNode, SchemeLink
from Orange.canvas.registry import cache
from unittest import TestCase

log = logging.getLogger(__name__)

app = QApplicationManager()


class OrangeWorflowTest(TestCase):
    """Define a specific TestCase reltive to OrangeWorkflow"""

    def addWidget(self, widgetID):
        """Create an instanciation of the widgetID

        :return SchemeNode: the node created
        """
        assert type(widgetID) is str
        assert self.widget_registry.has_widget(widgetID)
        widget_desc = self.widget_registry.widget(widgetID)
        myItem = SchemeNode(widget_desc)
        self.canvas_window.current_document().scheme().add_node(myItem)
        return myItem

    def link(self, schemeNode1, outNode1, schemeNode2, inputNode2):
        """Create a link between the node on the given input and output

        :param SchemeNode schemeNode1: the node emitting the output signal
        :param outNode1: the output chanel to be linked
        :param SchemeNode schemeNode2: the node receiving the input signal
        :param inputNode2: the input chanel to be linked
        :return: the link created between the two nodes
        """
        assert type(schemeNode1) is SchemeNode
        assert type(schemeNode2) is SchemeNode
        assert type(outNode1) is str
        assert type(inputNode2) is str
        link = SchemeLink(schemeNode1, outNode1, schemeNode2, inputNode2)
        self.canvas_window.current_document().scheme().add_link(link)
        return link

    def getWidgetForNode(self, node):
        """

        :param SchemeNode node: the node for which we want the corresponding
            widget
        :return: the widget instanciated for the given node
        """
        assert type(node) is SchemeNode
        return self.canvas_window.current_document().scheme().widget_for_node(node)

    @classmethod
    def setUpClass(cls):
        cls.init(cls)
        TestCase.setUpClass()

    @classmethod
    def tearDownClass(cls):
        cls.canvas_window.deleteLater()
        # make sure no events are resting
        while app.hasPendingEvents():
            app.processEvents()
            cls.processOrangeEvents(cls)
        app.flush()
        del cls.canvas_window

        gc.collect()
        # Collect any cycles before deleting the QApplication instance
        TestCase.tearDownClass()

    def tearDown(self):
        while app.hasPendingEvents():
            app.processEvents()
            TestCase.tearDown(self)

    @staticmethod
    def fix_osx_10_9_private_font():
        # Fix fonts on Os X (QTBUG 47206, 40833, 32789)
        if sys.platform == "darwin":
            import platform

            try:
                version = platform.mac_ver()[0]
                version = float(version[: version.rfind(".")])
                if version >= 10.11:  # El Capitan
                    QFont.insertSubstitution(".SF NS Text", "Helvetica Neue")
                elif version >= 10.10:  # Yosemite
                    QFont.insertSubstitution(
                        ".Helvetica Neue DeskInterface", "Helvetica Neue"
                    )
                elif version >= 10.9:
                    QFont.insertSubstitution(".Lucida Grande UI", "Lucida Grande")
            except AttributeError:
                pass

    @staticmethod
    def fix_win_pythonw_std_stream():
        """
        On windows when running without a console (using pythonw.exe) the
        std[err|out] file descriptors are invalid and start throwing exceptions
        when their buffer is flushed (`http://bugs.python.org/issue706263`_)

        """
        if (
            sys.platform == "win32"
            and os.path.basename(sys.executable) == "pythonw.exe"
        ):
            if sys.stdout is None:
                sys.stdout = open(os.devnull, "w")
            if sys.stderr is None:
                sys.stderr = open(os.devnull, "w")

    def processOrangeEvents(self):
        self.canvas_window.current_document().scheme().signal_manager.process_queued()

    def processOrangeEventsStack(self):
        node_update_front = (
            self.canvas_window.current_document()
            .scheme()
            .signal_manager.node_update_front()
        )
        if node_update_front:
            self.processOrangeEvents()
            self.processOrangeEventsStack()

    def init(self):
        # Fix streams before configuring logging (otherwise it will store
        # and write to the old file descriptors)
        OrangeWorflowTest.fix_win_pythonw_std_stream()

        # Try to fix fonts on OSX Mavericks
        OrangeWorflowTest.fix_osx_10_9_private_font()

        # File handler should always be at least INFO level so we need
        # the application root level to be at least at INFO.

        rootlogger = logging.getLogger(canvas.__name__)

        # Standard output stream handler at the requested level
        stream_hander = logging.StreamHandler()
        rootlogger.addHandler(stream_hander)

        log.info("Starting 'Orange Canvas' application.")
        log.debug("Starting CanvasApplicaiton ")
        # NOTE: config.init() must be called after the QApplication constructor
        config.init()

        file_handler = logging.FileHandler(
            filename=os.path.join(config.log_dir(), "canvas.log"), mode="w"
        )

        rootlogger.addHandler(file_handler)

        # intercept any QFileOpenEvent requests until the main window is
        # fully initialized.
        # NOTE: The QApplication must have the executable ($0) and filename
        # arguments passed in argv otherwise the FileOpen events are
        # triggered for them (this is done by Cocoa, but QApplicaiton filters
        # them out if passed in argv)

        open_requests = []

        def onrequest(url):
            log.info("Received an file open request %s", url)
            open_requests.append(url)

        app.fileOpenRequest.connect(onrequest)

        stylesheet = "orange"
        stylesheet_string = None

        if stylesheet != "none":
            if os.path.isfile(stylesheet):
                with open(stylesheet, "r") as f:
                    stylesheet_string = f.read()
            else:
                if not os.path.splitext(stylesheet)[1]:
                    # no extension
                    stylesheet = os.path.extsep.join([stylesheet, "qss"])

                pkg_name = canvas.__name__
                resource = "styles/" + stylesheet

                if pkg_resources.resource_exists(pkg_name, resource):
                    stylesheet_string = pkg_resources.resource_string(
                        pkg_name, resource
                    ).decode()

                    base = pkg_resources.resource_filename(pkg_name, "styles")

                    pattern = re.compile(
                        r"^\s@([a-zA-Z0-9_]+?)\s*:\s*([a-zA-Z0-9_/]+?);\s*$",
                        flags=re.MULTILINE,
                    )

                    matches = pattern.findall(stylesheet_string)

                    for prefix, search_path in matches:
                        QDir.addSearchPath(prefix, os.path.join(base, search_path))
                        log.info(
                            "Adding search path %r for prefix, %r", search_path, prefix
                        )

                    stylesheet_string = pattern.sub("", stylesheet_string)

                else:
                    log.info("%r style sheet not found.", stylesheet)

        # Add the default canvas_icons search path
        dirpath = os.path.abspath(os.path.dirname(canvas.__file__))
        QDir.addSearchPath("canvas_icons", os.path.join(dirpath, "icons"))
        self.canvas_window = CanvasMainWindow()
        self.canvas_window.setWindowIcon(config.application_icon())

        if stylesheet_string is not None:
            self.canvas_window.setStyleSheet(stylesheet_string)

        reg_cache = None
        widget_discovery = qt.QtWidgetDiscovery(cached_descriptions=reg_cache)

        self.widget_registry = qt.QtWidgetRegistry()

        widget_discovery.found_category.connect(self.widget_registry.register_category)
        widget_discovery.found_widget.connect(self.widget_registry.register_widget)

        cache_filename = os.path.join(cache_dir(), "widget-registry.pck")
        entry_points = []
        for entry_point in config.widgets_entry_points():
            if entry_point.name != "Orange Widgets":
                entry_points.append(entry_point)
        widget_discovery.run(entry_points)
        # Store cached descriptions
        cache.save_registry_cache(widget_discovery.cached_descriptions)
        with open(cache_filename, "wb") as f:
            pickle.dump(WidgetRegistry(self.widget_registry), f)

        set_global_registry(self.widget_registry)
        self.canvas_window.set_widget_registry(self.widget_registry)
        self.canvas_window.show()

        # Progress events to make sure the self.canvas_window layout has
        # a chance to activate (the welcome dialog is modal and will
        # block the event queue, plus we need a chance to receive open file
        # signals when running without a splash screen)
        app.processEvents()
        log.info("discovery ended")
        # Tee stdout and stderr into Output dock
        log_view = self.canvas_window.log_view()

        stdout = TextStream()
        stdout.stream.connect(log_view.write)
        if sys.stdout:
            stdout.stream.connect(sys.stdout.write)
            stdout.flushed.connect(sys.stdout.flush)
        stderr = TextStream()
        error_writer = log_view.formated(color=Qt.red)
        stderr.stream.connect(error_writer.write)
        if sys.stderr:
            stderr.stream.connect(sys.stderr.write)
            stderr.flushed.connect(sys.stderr.flush)

        log.info("End initialization")



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/


__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "08/08/2019"


import Orange.data
from est.core.types import XASObject
import numpy
import logging

_logger = logging.getLogger(__name__)


class Converter(object):
    """This converter insure a minimal conversion between xas_object and
    Orange.data.Table by only storing energy and absorbed beam (mu)"""

    @staticmethod
    def toXASObject(data_table):
        energy = _retrieve_energy(data_table.domain)
        mu = data_table.X
        mu = numpy.swapaxes(mu, 1, 0)
        mu = mu.reshape(mu.shape[0], mu.shape[1], 1)
        # note: for now we only consider 2D spectra...
        return XASObject(energy=energy, spectra=mu, dim2=1, dim1=mu.shape[-1])

    @staticmethod
    def toDataTable(xas_object):
        _logger.warning(
            "casting xas_object to Orange.data.Table might bring "
            "lost of some information (process flow, "
            "treatment result...). Only keep energy and absorbed "
            "beam information"
        )
        # TODO: prendre normalized_mu and normalized_energy if exists,
        # otherwise take mu and energy...
        spectra = xas_object.spectra.map_to(data_info="mu")
        # invert dimensions and axis to fit spectroscopy add-on
        X = spectra.reshape((spectra.shape[0], -1))
        X = numpy.swapaxes(X, 0, 1)

        domain = Orange.data.Domain(
            attributes=[
                Orange.data.ContinuousVariable.make("%f" % f) for f in xas_object.energy
            ]
        )
        data = Orange.data.Table.from_numpy(domain=domain, X=X)
        return data


def _retrieve_energy(domain):
    """
    Return x of the data. If all attribute names are numbers,
    return their values. If not, return indices.
    """
    energy = numpy.arange(len(domain.attributes), dtype="f")
    try:
        energy = numpy.array([float(a.name) for a in domain.attributes])
    except:
        _logger.error("fail to retrieve energy from attributes")
    return energy



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "06/07/2019"


import functools
import logging

from Orange.widgets import gui
from Orange.widgets.settings import Setting
from Orange.widgets.widget import OWWidget
from Orange.widgets.widget import Input, Output
import Orange.data
from est.core.types import XASObject
from PyMca5.PyMcaGui.physics.xas.XASPostEdgeParameters import XASPostEdgeParameters
from silx.gui import qt
from silx.gui.plot import LegendSelector

import est.core.process.pymca.exafs
from orangecontrib.est.process import _ProcessForOrangeMixIn, ProcessRunnable
from est.gui.XasObjectViewer import XasObjectViewer, ViewType
from est.gui.XasObjectViewer import (
    _exafs_signal_plot,
    _exafs_postedge_plot,
    _exafs_knots_plot,
)
from orangecontrib.est.progress import QProgress
from orangecontrib.est.utils import Converter
from ..container import _ParameterWindowContainer


_logger = logging.getLogger(__file__)


class ExafsWindow(qt.QMainWindow):
    def __init__(self, parent=None):
        qt.QMainWindow.__init__(self, parent)
        mapKeys = ["mu", "EXAFS.EXAFSKValues", "EXAFS.EXAFSSignal", "EXAFS.PostEdgeB"]
        self.xasObjViewer = XasObjectViewer(mapKeys=mapKeys)
        self.xasObjViewer._spectrumViews[0]._plotWidget.getXAxis().setLabel("K")
        self.xasObjViewer._spectrumViews[0]._plotWidget.getYAxis().setLabel(
            "Normalized Units"
        )
        self.setCentralWidget(self.xasObjViewer)

        # pymca window
        self._pymcaWindow = _ParameterWindowContainer(
            parent=self, parametersWindow=XASPostEdgeParameters
        )
        dockWidget = qt.QDockWidget(parent=self)
        dockWidget.setWidget(self._pymcaWindow)
        self.addDockWidget(qt.Qt.RightDockWidgetArea, dockWidget)
        dockWidget.setAllowedAreas(qt.Qt.RightDockWidgetArea | qt.Qt.LeftDockWidgetArea)
        dockWidget.setFeatures(qt.QDockWidget.NoDockWidgetFeatures)
        self.setWindowFlags(qt.Qt.Widget)

        # legend selector
        self.legendDockWidget = LegendSelector.LegendsDockWidget(
            parent=self, plot=self.xasObjViewer._spectrumViews[0]._plotWidget
        )
        self.legendDockWidget.setAllowedAreas(
            qt.Qt.RightDockWidgetArea | qt.Qt.LeftDockWidgetArea
        )
        self.legendDockWidget.setFeatures(qt.QDockWidget.NoDockWidgetFeatures)
        self.addDockWidget(qt.Qt.RightDockWidgetArea, self.legendDockWidget)

        # volume key selection
        self.addDockWidget(
            qt.Qt.RightDockWidgetArea, self.xasObjViewer._mapView.keySelectionDocker
        )

        # plot settings
        for ope in (_exafs_signal_plot, _exafs_postedge_plot, _exafs_knots_plot):
            self.xasObjViewer._spectrumViews[0].addCurveOperation(ope)

        self.setWindowFlags(qt.Qt.Widget)

        # connect signal / slot
        self.xasObjViewer.viewTypeChanged.connect(self._updateLegendView)

        # set up
        self._updateLegendView()

    def getNCurves(self):
        return len(self.xasObjViewer.getAllCurves())

    def _updateLegendView(self):
        index, viewType = self.xasObjViewer.getViewType()
        self.legendDockWidget.setVisible(viewType is ViewType.spectrum)
        self.xasObjViewer._mapView.keySelectionDocker.setVisible(
            viewType is ViewType.map
        )


class ExafsOW(_ProcessForOrangeMixIn, OWWidget):
    """
    Widget used for signal extraction
    """

    name = "exafs"
    id = "orange.widgets.est.pymca.exafs"
    description = "Progress signal extraction"
    icon = "icons/exafs.png"
    priority = 3
    category = "esrfWidgets"
    keywords = ["spectroscopy", "signal"]

    want_main_area = True
    resizing_enabled = True
    allows_cycle = False

    process_function = est.core.process.pymca.exafs.PyMca_exafs

    _larchSettings = Setting(dict())
    """Store the configuration of the PyMca XASClass"""

    class Inputs:
        xas_obj = Input("xas_obj", XASObject, default=True)
        # simple compatibility for some Orange widget and especialy the
        # 'spectroscopy add-on'
        data_table = Input("Data", Orange.data.Table)

    class Outputs:
        res_xas_obj = Output("xas_obj", XASObject)
        # by default we want to avoid sending 'Orange.data.Table' to avoid
        # loosing the XASObject flow process and results.

    def __init__(self):
        super().__init__()
        self._latest_xas_obj = None
        self._window = ExafsWindow()
        layout = gui.vBox(self.mainArea, "exafs").layout()
        layout.addWidget(self._window)

        self._window.xasObjViewer._spectrumViews[0]._plotWidget.getXAxis().setLabel("K")
        self._window.xasObjViewer._spectrumViews[0]._plotWidget.getYAxis().setLabel(
            "Normalized Units"
        )

        # manage settings
        if self._larchSettings != dict():
            self._window._pymcaWindow.setParameters(self._larchSettings)

        # signal / slot connection
        self._window._pymcaWindow.sigChanged.connect(self._updateProcess)

    def _updateProcess(self):
        self._update_settings()
        if self._latest_xas_obj:
            self.process(self._latest_xas_obj)

    def _update_settings(self):
        self._larchSettings = self._window._pymcaWindow.getParameters()

    @Inputs.data_table
    def processFrmDataTable(self, data_table):
        if data_table is None:
            return
        self.process(Converter.toXASObject(data_table=data_table))

    @Inputs.xas_obj
    def process(self, xas_obj):
        if xas_obj is None:
            return

        if not self._canProcess():
            _logger.warning(
                "There is some processing on going already, will"
                "not process the new dataset"
            )

        self._latest_xas_obj = xas_obj
        self._startProcess()

        # setup the exafs process
        process_obj = QPyMca_exafs()
        process_obj._advancement.sigProgress.connect(self._setProgressValue)
        process_obj.set_properties(
            {"_pymcaSettings": self._window._pymcaWindow.getParameters()}
        )

        # update the processing thread
        thread = self.getProcessingThread()
        thread.init(process_obj=process_obj, xas_obj=self._latest_xas_obj)
        self._callback_finish = functools.partial(
            self._endProcess, self._latest_xas_obj
        )
        thread.finished.connect(self._callback_finish)
        # start processing
        thread.start(priority=qt.QThread.LowPriority)

    def _setProgressValue(self, value):
        self._progress.widget.progressBarSet(value)


class QPyMca_exafs(est.core.process.pymca.exafs.PyMca_exafs):
    """
    Normalization able to give advancement using qt.Signal and QThreadPool
    """

    def __init__(self):
        est.core.process.pymca.exafs.PyMca_exafs.__init__(self)
        self._advancement = QProgress("normalization")

    def _pool_process(self, xas_obj):
        self.pool = qt.QThreadPool()
        self.pool.setMaxThreadCount(5)
        for spectrum in xas_obj.spectra:
            runnable = ProcessRunnable(
                fct=est.core.process.pymca.exafs.process_spectr_exafs,
                spectrum=spectrum,
                configuration=xas_obj.configuration,
                callback=self._advancement.increaseAdvancement,
            )
            self.pool.start(runnable)
        self.pool.waitForDone()



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "06/11/2019"

from silx.io.spech5 import SpecFile
from est.units import ur
import functools
import numpy


@functools.lru_cache(maxsize=2)
def read_spectrum(
    spec_file,
    energy_col_name=None,
    absorption_col_name=None,
    monitor_col_name=None,
    energy_unit=ur.eV,
    scan_header_S=None,
):
    """
    :note: the cache is sued because we call twice the same function for energy
           and absorption


    :param str spec_file: path to the spec file containing the spectra
                          definition
    :param str energy_col_index:
    :param str absorption_col_index:
    :param Union[str,None] monitor_col_name: name of the column to get monitor
    :param Union[str,None] scan_header_S: name of the scan to consider

    :return: (energy, mu)
    :rtype: tuple energy and list of absorption
    """
    spec_file = SpecFile(spec_file)
    energy = None
    mu = None
    for i_data, scan in enumerate(spec_file):
        # if a scan header 'title' is provided
        if scan_header_S is not None and scan_header_S != scan.scan_header_dict["S"]:
            continue

        if (
            energy_col_name not in scan.labels
            and absorption_col_name not in scan.labels
        ):
            continue

        # get energy
        if energy_col_name is not None:
            energy = spec_file.data_column_by_name(
                scan_index=i_data, label=energy_col_name
            )
            if energy is not None:
                energy = (energy * energy_unit).m_as(ur.eV)

        # get absorption
        if absorption_col_name is not None:
            mu = spec_file.data_column_by_name(
                scan_index=i_data, label=absorption_col_name
            )

            if energy is not None:
                assert len(mu) == len(
                    energy
                ), "different number of elements between energy {} and absorption {}".format(
                    len(energy), len(mu)
                )

        # get monitor
        if monitor_col_name is not None:
            monitor = spec_file.data_column_by_name(
                scan_index=i_data, label=monitor_col_name
            )
            mu = mu / monitor

        return energy, numpy.asarray(mu)
    return energy, mu



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2017-2019 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "06/26/2019"


import unittest
import urllib
import tempfile
import os
import shutil
from est.core.types import Spectrum, XASObject
from silx.io.utils import h5py_read_dataset
import h5py

try:
    from est.io.utils.larch import read_ascii
    from est.core.process.larch.pre_edge import Larch_pre_edge
except ImportError:
    has_larch = False
else:
    has_larch = True


@unittest.skipIf(has_larch is False, "xraylarch not installed")
class TestWriteProcess(unittest.TestCase):
    """
    Insure saving processes works
    """

    def setUp(self):
        self.outputdir = tempfile.mkdtemp()
        xmu_url = "https://raw.githubusercontent.com/xraypy/xraylarch/master/examples/xafs/cu_rt01.xmu"
        self.data_file = os.path.join(self.outputdir, "cu_rt01.xmu")

        with urllib.request.urlopen(xmu_url) as response, open(
            self.data_file, "wb"
        ) as out_file:
            data = response.read()  # a `bytes` object
            out_file.write(data)

        assert os.path.exists(self.data_file)
        energy, mu = read_ascii(self.data_file)
        self.spectrum = Spectrum(energy=energy, mu=mu)
        self.xas_obj = XASObject(spectra=(self.spectrum,), energy=self.spectrum.energy)
        self.output_file = os.path.join(self.outputdir, "output.h5")

    def tearDown(self):
        shutil.rmtree(self.outputdir)

    def testProcess(self):
        self.assertTrue(self.spectrum.pre_edge is None)
        self.assertTrue(self.spectrum.e0 is None)
        process = Larch_pre_edge()
        process.process(self.xas_obj)
        self.assertTrue(self.spectrum.pre_edge is not None)
        self.assertTrue(self.spectrum.e0 is not None)
        # check process
        self.xas_obj.dump(self.output_file)
        self.assertTrue(os.path.exists(self.output_file))
        with h5py.File(self.output_file, "r") as h5f:
            scan = h5f["scan1"]
            pre_edge_process = scan["xas_process_1"]
            # check general informqtion
            self.assertTrue("class_instance" in pre_edge_process)
            self.assertTrue("date" in pre_edge_process)
            self.assertTrue("processing_order" in pre_edge_process)
            self.assertTrue("program" in pre_edge_process)
            self.assertTrue("version" in pre_edge_process)
            self.assertEqual(
                h5py_read_dataset(pre_edge_process["program"]), "larch_pre_edge"
            )
            # check results
            self.assertTrue("results" in pre_edge_process)
            results_grp = pre_edge_process["results"]
            self.assertTrue("Mu" in results_grp)
            mu_grp = results_grp["Mu"]
            self.assertTrue("units" in mu_grp.attrs)
            self.assertTrue("units_latex" in mu_grp.attrs)

            # check plots
            self.assertTrue("plots" in pre_edge_process)
            plot_0_grp = pre_edge_process["plots"]["plot_0"]
            self.assertTrue("Mu" in plot_0_grp)
            self.assertTrue("energy" in plot_0_grp)
            self.assertTrue("units" in plot_0_grp["Mu"].attrs)
            self.assertTrue("units_latex" in plot_0_grp["Mu"].attrs)
            self.assertTrue("title" in plot_0_grp.attrs)
            self.assertTrue("title_latex" in plot_0_grp.attrs)


def suite():
    test_suite = unittest.TestSuite()
    for ui in (TestWriteProcess,):
        test_suite.addTest(unittest.defaultTestLoader.loadTestsFromTestCase(ui))
    return test_suite


if __name__ == "__main__":
    unittest.main(defaultTest="suite")



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "06/11/2019"

from est.core.types import XASObject
from est.core.types import Spectrum
from est.core.types import Spectra
from est.core.process.process import _input_desc
from est.core.process.process import _output_desc
from .process import Process
import logging
import numpy
import pint
from typing import Union

_logger = logging.getLogger(__name__)


def xas_roi(xas_obj):
    """
    apply roi on the XASObject.spectra

    :param xas_obj: object containing the configuration and spectra to process
    :type: Union[:class:`.XASObject`, dict]
    :return: spectra dict
    :rtype: :class:`.XASObject`
    """
    xas_roi_process = ROIProcess()
    return xas_roi_process.process(xas_obj=xas_obj)


class _ROI(object):
    def __init__(self, origin, size):
        self.origin = origin
        self.size = size

    @property
    def origin(self) -> tuple:
        return self.__origin

    @origin.setter
    def origin(self, origin: Union[list, tuple]):
        self.__origin = int(origin[0]), int(origin[1])

    @property
    def size(self) -> tuple:
        return self.__size

    @size.setter
    def size(self, size: Union[list, tuple]):
        self.__size = int(size[0]), int(size[1])

    def to_dict(self) -> dict:
        return {"origin": self.origin, "size": self.size}

    @staticmethod
    def from_dict(ddict: dict):
        return _ROI(origin=ddict["origin"], size=ddict["size"])

    @staticmethod
    def from_silx_def(silx_roi):
        origin = silx_roi.getOrigin()
        size = silx_roi.getSize()
        return _ROI(origin=origin, size=size)


class ROIProcess(Process):

    inputs = [
        _input_desc(name="xas_obj", type=XASObject, handler="process", doc=""),
    ]

    outputs = [
        _output_desc(name="xas_obj", type=XASObject, doc=""),
    ]

    def __init__(self):
        Process.__init__(self, name="roi")
        self._roi = None

    def set_properties(self, properties: dict):
        if "roi" in properties:
            self._roi = _ROI.from_dict(properties["roi"])

    def setRoi(self, origin: Union[list, tuple], size: Union[list, tuple]):
        self._roi = _ROI(origin=origin, size=size)

    def _apply_roi(self, xas_obj, roi):
        _logger.warning("applying roi")
        assert isinstance(xas_obj, XASObject)
        assert isinstance(roi, _ROI)
        assert type(roi.origin[0]) is int
        assert type(roi.origin[1]) is int
        assert type(roi.size[0]) is int
        assert type(roi.size[1]) is int
        ymin, ymax = roi.origin[1], roi.origin[1] + roi.size[1]
        xmin, xmax = roi.origin[0], roi.origin[0] + roi.size[0]
        # clip roi
        ymin = max(ymin, 0)
        ymax = min(ymax, xas_obj.spectra.shape[0])
        xmin = max(xmin, 0)
        xmax = min(xmax, xas_obj.spectra.shape[1])
        assert type(ymin) is int
        assert type(ymax) is int
        assert type(xmin) is int
        assert type(xmax) is int
        assert ymax - ymin == roi.size[1]
        assert xmax - xmin == roi.size[0]

        # TODO: create a spectra object which deal automatically with
        # the following. Should be part of spectra object
        volumes = {}
        for key in xas_obj.spectra_keys():
            if isinstance(
                xas_obj.spectra.data.flat[0][key], (numpy.ndarray, pint.Quantity)
            ):
                # there is no processing for the _larch_grp_members case
                if key == "_larch_grp_members":
                    continue
                volume = xas_obj.spectra.map_to(data_info=key)
                volume_res = volume[:, ymin:ymax, xmin:xmax]
                volume_res = volume_res.reshape(volume_res.shape[0], -1)
                volumes[key] = volume_res

        # then rewrite spectrum list
        assert "Mu" in volumes
        assert "Energy" in volumes
        _mus = volumes["Mu"]
        _energies = volumes["Energy"]
        new_spectra = []
        assert _mus.shape == _energies.shape
        n_new_spectrum = roi.size[0] * roi.size[1]
        assert n_new_spectrum == _mus.shape[1]
        # create the new spectrum
        for i_new_spectrum in range(n_new_spectrum):
            new_spectra.append(
                Spectrum(
                    mu=_mus[:, i_new_spectrum], energy=_energies[:, i_new_spectrum]
                )
            )

        # update them from extra keys
        del volumes["Mu"]
        del volumes["Energy"]
        for key in volumes:
            for slice, new_spectrum in zip(volumes[key], new_spectra):
                new_spectrum[key] = slice

        # set the spectra
        xas_obj.spectra = Spectra(energy=_energies[:, 0], spectra=new_spectra)

        # update the dimensions
        xas_obj.spectra.reshape((roi.size[1], roi.size[0]))
        return xas_obj

    def process(self, xas_obj: Union[dict, XASObject]):
        """

        :param xas_obj: object containing the configuration and spectra to process
        :type: Union[:class:`.XASObject`, dict]
        :return: spectra dict
        :rtype: :class:`.XASObject`
        """
        assert xas_obj is not None
        _xas_obj = self.getXasObject(xas_obj=xas_obj)

        # existing roi is priority. This is the case if called from pushworkflow
        # for example.
        if self._roi is not None:
            return self._apply_roi(xas_obj=_xas_obj, roi=self._roi)
        elif "roi" in _xas_obj.configuration:
            roi_dict = _xas_obj.configuration["roi"]
            origin, size = roi_dict["origin"], roi_dict["size"]
            return self._apply_roi(xas_obj=_xas_obj, roi=_ROI(origin, size))
        else:
            return _xas_obj

    __call__ = process
